#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from mock import patch
import unittest

from oslo_config import cfg

from gbpservice.nfp.common import constants as nfp_constants
from gbpservice.nfp.common import exceptions
from gbpservice.nfp.orchestrator.drivers import (
    orchestration_driver
)

import uuid as pyuuid

cfg.CONF.import_group('keystone_authtoken', 'keystonemiddleware.auth_token')
OPENSTACK_DRIVER_CLASS_PATH = ('gbpservice.nfp.orchestrator'
                               '.openstack.openstack_driver')
NFP_GBP_NETWORK_DRIVER_CLASS_PATH = ('gbpservice.nfp.orchestrator'
                                     '.coal.networking'
                                     '.nfp_gbp_network_driver')
NFP_NEUTRON_NETWORK_DRIVER_CLASS_PATH = ('gbpservice.nfp.orchestrator'
                                         '.coal.networking'
                                         '.nfp_neutron_network_driver')


@patch(OPENSTACK_DRIVER_CLASS_PATH + '.KeystoneClient.__init__',
       mock.MagicMock(return_value=None))
@patch(OPENSTACK_DRIVER_CLASS_PATH + '.NovaClient.__init__',
       mock.MagicMock(return_value=None))
@patch(NFP_GBP_NETWORK_DRIVER_CLASS_PATH + '.NFPGBPNetworkDriver.__init__',
       mock.MagicMock(return_value=None))
@patch(NFP_NEUTRON_NETWORK_DRIVER_CLASS_PATH +
       '.NFPNeutronNetworkDriver.__init__',
       mock.MagicMock(return_value=None))
class OrchestrationDriverTestCase(unittest.TestCase):

    def test_create_network_function_device(self):
        driver = orchestration_driver.OrchestrationDriver(
            cfg.CONF,
            supports_device_sharing=True,
            supports_hotplug=True,
            max_interfaces=8)
        driver.network_handler = driver.network_handlers['gbp']

        # Mock the client methods
        driver.identity_handler.get_admin_token = mock.MagicMock(
            return_value='token')
        driver.identity_handler.get_tenant_id = mock.MagicMock(
            return_value='8')
        driver.identity_handler.get_keystone_creds = mock.MagicMock(
            return_value=(None, None, 'admin', None))
        driver.network_handler.create_port = mock.MagicMock(
            return_value={'id': str(pyuuid.uuid4()),
                          'port_id': str(pyuuid.uuid4())})
        driver.network_handler.set_promiscuos_mode = mock.MagicMock(
            return_value=None)
        driver.network_handler.set_promiscuos_mode_fast = mock.MagicMock(
            return_value=None)
        driver.compute_handler_nova.get_image_id = mock.MagicMock(
            return_value='6')
        driver.compute_handler_nova.get_image_metadata = mock.MagicMock(
            return_value=[])
        driver.compute_handler_nova.create_instance = mock.MagicMock(
            return_value='8')
        driver.network_handler.delete_port = mock.MagicMock(
            return_value=None)
        driver.network_handler.get_port_id = mock.MagicMock(return_value='7')
        driver.network_handler.get_port_details = mock.MagicMock(
            return_value=('a.b.c.d',
                          'aa:bb:cc:dd:ee:ff',
                          'p.q.r.s/t',
                          'w.x.y.z'))
        driver.network_handler.get_neutron_port_details = mock.MagicMock(
            return_value=(1, 2, 3, 4,
                          {'port': {}},
                          {'subnet': {}}))

        # test for create device when interface hotplug is enabled
        device_data = {'service_details': {'device_type': 'xyz',
                                           'service_type': 'firewall',
                                           'service_vendor': 'vyos',
                                           'network_mode': 'gbp'},
                       'name': 'FIREWALL.vyos.1.2',
                       'volume_support': None,
                       'volume_size': None,
                       'management_network_info': {'id': '2'},
                       'ports': [{'id': '3',
                                  'port_model': 'gbp',
                                  'port_classification': 'provider'},
                                 {'id': '4',
                                  'port_model': 'gbp',
                                  'port_classification': 'consumer'}],
                       'token': str(pyuuid.uuid4()),
                       'admin_tenant_id': str(pyuuid.uuid4())}
        self.assertRaises(exceptions.ComputePolicyNotSupported,
                          driver.create_network_function_device,
                          device_data)
        device_data['service_details']['device_type'] = 'nova'
        self.assertIsInstance(driver.create_network_function_device(
            device_data),
            dict,
            msg=('Return value from the'
                 ' create_network_function_device call'
                 ' is not a dictionary'))

        # test for create device along with provider port
        driver.supports_hotplug = False
        self.assertIsInstance(driver.create_network_function_device(
            device_data),
            dict,
            msg=('Return value from the'
                 ' create_network_function_device call'
                 ' is not a dictionary'))

    def test_delete_network_function_device(self):
        driver = orchestration_driver.OrchestrationDriver(
            cfg.CONF,
            supports_device_sharing=True,
            supports_hotplug=True,
            max_interfaces=8)
        driver.network_handler = driver.network_handlers['gbp']

        # Mock the client methods
        driver.identity_handler.get_admin_token = mock.MagicMock(
            return_value='token')
        driver.identity_handler.get_tenant_id = mock.MagicMock(
            return_value='8')
        driver.identity_handler.get_keystone_creds = mock.MagicMock(
            return_value=(None, None, 'admin', None))
        driver.compute_handler_nova.delete_instance = mock.MagicMock(
            return_value=None)
        driver.network_handler.delete_port = mock.MagicMock(return_value=None)

        device_data = {'id': '1',
                       'service_details': {'device_type': 'xyz',
                                           'service_type': 'firewall',
                                           'service_vendor': 'vyos',
                                           'network_mode': 'gbp'},
                       'mgmt_port_id': {'id': '3',
                                        'port_model': 'gbp',
                                        'port_classification': 'mgmt'}}
        self.assertRaises(exceptions.ComputePolicyNotSupported,
                          driver.delete_network_function_device,
                          device_data)
        device_data['service_details']['device_type'] = 'nova'
        self.assertIsNone(driver.delete_network_function_device(device_data))

    def test_get_network_function_device_status(self):
        driver = orchestration_driver.OrchestrationDriver(
            cfg.CONF,
            supports_device_sharing=True,
            supports_hotplug=True,
            max_interfaces=8)

        # Mock the client methods
        driver.identity_handler.get_admin_token = mock.MagicMock(
            return_value='token')
        driver.identity_handler.get_tenant_id = mock.MagicMock(
            return_value='8')
        driver.identity_handler.get_keystone_creds = mock.MagicMock(
            return_value=(None, None, 'admin', None))
        driver.compute_handler_nova.get_instance = mock.MagicMock(
            return_value={'status': 'ACTIVE'})

        device_data = {'id': '1',
                       'service_details': {'device_type': 'xyz',
                                           'service_type': 'firewall',
                                           'service_vendor': 'vyos',
                                           'network_mode': 'gbp'},
                       'token': str(pyuuid.uuid4()),
                       'tenant_id': str(pyuuid.uuid4())}

        self.assertRaises(exceptions.ComputePolicyNotSupported,
                          driver.get_network_function_device_status,
                          device_data)
        device_data['service_details']['device_type'] = 'nova'

        self.assertTrue(
            driver.get_network_function_device_status(device_data) ==
            'ACTIVE')

    def test_plug_network_function_device_interfaces(self):
        driver = orchestration_driver.OrchestrationDriver(
            cfg.CONF,
            supports_device_sharing=True,
            supports_hotplug=False,
            max_interfaces=8)
        driver.network_handler = driver.network_handlers['gbp']
        # Mock the client methods
        driver.identity_handler.get_admin_token = mock.MagicMock(
            return_value='token')
        driver.identity_handler.get_tenant_id = mock.MagicMock(
            return_value='8')
        driver.identity_handler.get_keystone_creds = mock.MagicMock(
            return_value=(None, None, 'admin', None))
        driver.network_handler.set_promiscuos_mode = mock.MagicMock(
            return_value=None)
        driver.network_handler.set_promiscuos_mode_fast = mock.MagicMock(
            return_value=None)
        driver.compute_handler_nova.attach_interface = mock.MagicMock(
            return_value=None)
        driver.compute_handler_nova.get_image_metadata = mock.MagicMock(
            return_value={})
        driver.network_handler.get_port_id = mock.MagicMock(return_value='7')

        device_data = {'id': '1',
                       'service_details': {'device_type': 'xyz',
                                           'service_type': 'firewall',
                                           'service_vendor': 'vyos',
                                           'network_mode': 'gbp'},
                       'ports': [{'id': '3',
                                  'port_model': 'gbp',
                                  'port_classification': 'provider'},
                                 {'id': '4',
                                  'port_model': 'neutron',
                                  'port_classification': 'consumer'}],
                       'provider_metadata': {},
                       'token': str(pyuuid.uuid4()),
                       'tenant_id': str(pyuuid.uuid4())}

        device_data['service_details']['device_type'] = 'nova'

        self.assertTrue(driver.plug_network_function_device_interfaces(
            device_data),
            msg='')

    def test_unplug_network_function_device_interfaces(self):
        driver = orchestration_driver.OrchestrationDriver(
            cfg.CONF,
            supports_device_sharing=True,
            supports_hotplug=False,
            max_interfaces=8)
        driver.network_handler = driver.network_handlers['gbp']

        driver.identity_handler.get_admin_token = mock.MagicMock(
            return_value='token')
        driver.identity_handler.get_tenant_id = mock.MagicMock(
            return_value='8')
        driver.identity_handler.get_keystone_creds = mock.MagicMock(
            return_value=(None, None, 'admin', None))
        driver.compute_handler_nova.detach_interface = mock.MagicMock(
            return_value=None)
        driver.compute_handler_nova.get_image_metadata = mock.MagicMock(
            return_value={})
        driver.network_handler.get_port_id = mock.MagicMock(return_value='7')

        device_data = {'id': '1',
                       'tenant_id': 'tenant_id',
                       'service_details': {'device_type': 'xyz',
                                           'service_type': 'firewall',
                                           'service_vendor': 'vyos',
                                           'network_mode': 'gbp'},
                       'provider_metadata': {},
                       'ports': [{'id': '3',
                                  'port_model': 'gbp',
                                  'port_classification': 'provider'},
                                 {'id': '4',
                                  'port_model': 'neutron',
                                  'port_classification': 'consumer'}]}

        device_data['service_details']['device_type'] = 'nova'

        self.assertTrue(driver.unplug_network_function_device_interfaces(
            device_data),
            msg='')

    def test_get_network_function_device_healthcheck_info(self):
        driver = orchestration_driver.OrchestrationDriver(
            cfg.CONF,
            supports_device_sharing=True,
            supports_hotplug=False,
            max_interfaces=8)

        device_data = {'id': '1',
                       'mgmt_ip_address': 'a.b.c.d'}

        self.assertIsInstance(
            driver.get_network_function_device_config(device_data,
                nfp_constants.HEALTHMONITOR_RESOURCE),
            dict, msg='')

    def test_get_network_function_device_config(self):
        driver = orchestration_driver.OrchestrationDriver(
            cfg.CONF,
            supports_device_sharing=True,
            supports_hotplug=False,
            max_interfaces=8)
        driver.network_handler = driver.network_handlers['gbp']

        driver.identity_handler.get_admin_token = mock.MagicMock(
            return_value='token')
        driver.network_handler.get_port_details = mock.MagicMock(
            return_value=('a.b.c.d',
                          'aa:bb:cc:dd:ee:ff',
                          'p.q.r.s/t',
                          'w.x.y.z'))

        device_data = {'service_details': {'device_type': 'xyz',
                                           'service_type': 'firewall',
                                           'service_vendor': 'vyos',
                                           'network_mode': 'gbp'},

                       'mgmt_ip_address': 'a.b.c.d',
                       'ports': [{'id': '3',
                                  'port_model': 'gbp',
                                  'port_classification': 'provider'}]}

        reply = driver.get_network_function_device_config(device_data,
                nfp_constants.GENERIC_CONFIG)
        self.assertIsInstance(reply, dict, msg='')
        self.assertTrue('config' in reply)



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.common import constants as q_const
from neutron import context as n_ctx
from neutron.extensions import l3
from neutron import manager

from apic_ml2.neutron.services.l3_router import apic_driver_api


class ApicGBPL3Driver(apic_driver_api.ApicL3DriverBase):

    def __init__(self, plugin):
        super(ApicGBPL3Driver, self).__init__()
        self._plugin = plugin
        self._apic_gbp = None

    @property
    def apic_gbp(self):
        if not self._apic_gbp:
            self._apic_gbp = manager.NeutronManager.get_service_plugins()[
                'GROUP_POLICY'].policy_driver_manager.policy_drivers[
                'apic'].obj
        return self._apic_gbp

    def _get_port_id_for_router_interface(self, context, router_id, subnet_id):
        filters = {'device_id': [router_id],
                   'device_owner': [q_const.DEVICE_OWNER_ROUTER_INTF],
                   'fixed_ips': {'subnet_id': [subnet_id]}}
        ports = self._plugin._core_plugin.get_ports(context.elevated(),
                                                    filters=filters)
        return ports[0]['id']

    def _update_router_gw_info(self, context, router_id, info, router=None):
        super(ApicGBPL3Driver, self)._update_router_gw_info(
            context, router_id, info, router)
        if info and 'network_id' in info:
            filters = {'device_id': [router_id],
                       'device_owner': [q_const.DEVICE_OWNER_ROUTER_GW],
                       'network_id': [info['network_id']]}
            ports = self._plugin._core_plugin.get_ports(context.elevated(),
                                                        filters=filters)
            self._plugin._core_plugin.update_port_status(
                context, ports[0]['id'], q_const.PORT_STATUS_ACTIVE)

    def add_router_interface_postcommit(self, context, router_id,
                                        interface_info):
        if 'subnet_id' in interface_info:
            port_id = self._get_port_id_for_router_interface(
                context, router_id, interface_info['subnet_id'])
        else:
            port_id = interface_info['port_id']

        self._plugin._core_plugin.update_port_status(context,
            port_id, q_const.PORT_STATUS_ACTIVE)

    def remove_router_interface_precommit(self, context, router_id,
                                          interface_info):
        if 'subnet_id' in interface_info:
            port_id = self._get_port_id_for_router_interface(
                context, router_id, interface_info['subnet_id'])
        else:
            port_id = interface_info['port_id']

        self._plugin._core_plugin.update_port_status(context,
            port_id, q_const.PORT_STATUS_DOWN)

    # Floating IP API
    def create_floatingip_precommit(self, context, floatingip):
        fip = floatingip['floatingip']
        tenant_id = self._plugin._get_tenant_id_for_create(context, fip)
        if self.apic_gbp:
            context.nat_pool_list = []
            for nat_pool in self.apic_gbp.nat_pool_iterator(context,
                    tenant_id, floatingip):
                context.nat_pool_list.append(nat_pool)

    def create_floatingip_postcommit(self, context, floatingip):
        port_id = floatingip.get('floatingip', {}).get('port_id')
        self._notify_port_update(port_id)
        if getattr(context, 'result', None):
            context.result['status'] = self._update_floatingip_status(
                context, context.result['id'])

    def update_floatingip_precommit(self, context, id, floatingip):
        port_id = self._get_port_mapped_to_floatingip(context, id)
        context.port_id_list = [port_id]

    def update_floatingip_postcommit(self, context, id, floatingip):
        port_id_list = getattr(context, 'port_id_list', [])
        port_id_list.append(
            floatingip.get('floatingip', {}).get('port_id'))
        for p in port_id_list:
            self._notify_port_update(p)
        status = self._update_floatingip_status(context, id)
        if getattr(context, 'result', None):
            context.result['status'] = status

    def delete_floatingip_precommit(self, context, id):
        port_id_list = [self._get_port_mapped_to_floatingip(context, id)]
        context.port_id_list = port_id_list

    def delete_floatingip_postcommit(self, context, id):
        self._notify_port_update(context.port_id_list[0])

    def _notify_port_update(self, port_id):
        context = n_ctx.get_admin_context()
        if self.apic_gbp and port_id:
            self.apic_gbp._notify_port_update(context, port_id)
            ptg, _ = self.apic_gbp._port_id_to_ptg(context, port_id)
            if ptg:
                self.apic_gbp._notify_head_chain_ports(ptg['id'])

    def _update_floatingip_status(self, context, fip_id):
        status = q_const.FLOATINGIP_STATUS_DOWN
        try:
            fip = self._plugin.get_floatingip(context, fip_id)
            if fip.get('port_id'):
                status = q_const.FLOATINGIP_STATUS_ACTIVE
            self._plugin.update_floatingip_status(context, fip_id, status)
        except l3.FloatingIPNotFound:
            pass
        return status

    def _get_port_mapped_to_floatingip(self, context, fip_id):
        try:
            fip = self._plugin.get_floatingip(context, fip_id)
            return fip.get('port_id')
        except l3.FloatingIPNotFound:
            pass
        return None



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db import model_base
import sqlalchemy as sa


class ApicReuseBdDB(model_base.BASEV2):
    __tablename__ = 'gp_apic_mapping_reuse_bds'
    l2_policy_id = sa.Column(
        sa.String(36), sa.ForeignKey('gp_l2_policies.id',
                                     ondelete='CASCADE'), primary_key=True)
    target_l2_policy_id = sa.Column(
        sa.String(36), sa.ForeignKey('gp_l2_policies.id'), nullable=False)


class ApicReuseBdDBMixin(object):

    def get_reuse_bd_l2policy(self, session, l2_policy_id):
        row = (session.query(ApicReuseBdDB).filter_by(
               l2_policy_id=l2_policy_id).first())
        return row

    def add_reuse_bd_l2policy(self, session, l2_policy_id,
                              target_l2_policy_id):
        with session.begin(subtransactions=True):
            row = ApicReuseBdDB(l2_policy_id=l2_policy_id,
                                target_l2_policy_id=target_l2_policy_id)
            session.add(row)

    def is_reuse_bd_target(self, session, l2_policy_id):
        return (session.query(ApicReuseBdDB).filter_by(
                target_l2_policy_id=l2_policy_id).first() is not None)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():

    # Run the Testcase:
    test = test_gbp_pa_neg()
    if test.test_pa_invalid_act_type() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PA_NEG_1')
    if test.test_pa_valid_type_inval_val() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PA_NEG_2')
    if test.test_pa_invalid_act_value() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PA_NEG_3')
    if test.test_pa_update_act_type() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PA_NEG_4')
    if test.test_pa_update_invalid_act_val() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PA_NEG_5')
    if test.test_pa_delete_invalid_pa() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PA_NEG_6')
    test.cleanup()
    utils_libs.report_results('test_gbp_pa_neg', 'test_results.txt')
    sys.exit(1)


class test_gbp_pa_neg(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    cmd = 'rm /tmp/test_gbp_pa_neg.log'
    commands.getoutput(cmd)
    hdlr = logging.FileHandler('/tmp/test_gbp_pa_neg.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self):
        """
        Init def
        """
        self._log.info("\n## START OF GBP POLICY_ACTION NEGATIVE TESTSUITE\n")
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.act_name = 'demo_act'

    def cleanup(self, tc_name=''):
        if tc_name != '':
            self._log.info('%s: FAILED' % (tc_name))
        for obj in ['ruleset', 'rule', 'classifier', 'action']:
            self.gbpcfg.gbp_del_all_anyobj(obj)

    def test_pa_invalid_act_type(self):

        self._log.info(
            "\n###################################################\n"
            "TESTCASE_GBP_PA_NEG_1: INVALID Action TYPE\n"
            "TEST_STEPS::\n"
            "Create and Verify Policy Action Errors Out for Invalid "
            "Action Type as Attr\n"
            "Invalid Action Type: Null string and string "
            "!= 'allow'/'redirect'\n"
            "###################################################\n")
        for _type in ["", "INVALID"]:
            if self.gbpcfg.gbp_action_config(
                    1, self.act_name, action_type=_type) != 0:
                self._log.info(
                    "\n## Step 1: Create Action with invalid Action Type=%s "
                    "did NOT Fail" %
                    (_type))
                return 0
            if self.gbpverify.gbp_action_verify(1, self.act_name) != 0:
                self._log.info(
                    "\n## Step 1A: Rollback of invalid Action create Failed")
                return 0
        self._log.info("\n## TESTCASE_GBP_PA_NEG_1: PASSED")

    def test_pa_valid_type_inval_val(self):

        self._log.info(
            "\n###################################################\n"
            "TESTCASE_GBP_PA_NEG_2: INVALID Action VALUE for "
            "VALID act_type \n"
            "TEST_STEPS::\n"
            "Create and Verify Policy Action Errors Out for Invalid "
            "Action Value for type=REDIRECT\n"
            "###################################################\n")
        if self.gbpcfg.gbp_action_config(
                1,
                self.act_name,
                action_type='redirect',
                action_value='INVALID') != 0:
            self._log.info(
                "\n## Step 1: Create Action with invalid Action Value "
                "did NOT Fail")
            return 0
        if self.gbpverify.gbp_action_verify(1, self.act_name) != 0:
            self._log.info(
                "\n## Step 1A: Rollback of Invalid Action create Failed")
            return 0
        self._log.info("\n## TESTCASE_GBP_PA_NEG_2: PASSED")

    def test_pa_invalid_act_value(self):

        self._log.info(
            "\n###################################################\n"
            "TESTCASE_GBP_PA_NEG_3: INVALID Action VALUE \n"
            "TEST_STEPS::\n"
            "Create and Verify Policy Action Errors Out for Invalid "
            "Action Value\n"
            "###################################################\n")
        if self.gbpcfg.gbp_action_config(
                1, self.act_name, action_value='INVALID') != 0:
            self._log.info(
                "\n## Step 1: Create Action with invalid Action Value "
                "did NOT Fail")
            return 0
        if self.gbpverify.gbp_action_verify(1, self.act_name) != 0:
            self._log.info(
                "\n## Step 1A: Rollback of Invalid Action create Failed")
            return 0
        self._log.info("\n## TESTCASE_GBP_PA_NEG_3: PASSED")

    def test_pa_update_act_type(self):

        self._log.info(
            "\n###################################################\n"
            "TESTCASE_GBP_PA_NEG_4: UPDATE Immutable ATTR act_type\n"
            "TEST_STEPS::\n"
            "Create a valid Policy Action with default action type\n"
            "Update name and act_type and Verify that it has failed to "
            "update the Policy Action\n"
            "###################################################\n")
        act_uuid = self.gbpcfg.gbp_action_config(1, self.act_name)
        if act_uuid == 0:
            self._log.info("## Step 1: Create Action == Failed")
            return 0
        if self.gbpcfg.gbp_action_config(
                2,
                act_uuid,
                name='grppol_act',
                action_type='redirect') != 0:
            self._log.info(
                "\n##Step 2: Updating Policy Action's Attrs name & "
                "action_type did NOT Fail")
            return 0
        if self.gbpverify.gbp_action_verify(
                1,
                self.act_name,
                id=act_uuid,
                action_type='allow',
                shared='False') == 0:
            self._log.info(
                "\n## Step 2B: Verify Policy Action Attrs are NOT "
                "updated == Failed")
            return 0
        self._log.info("\n## TESTCASE_GBP_PA_NEG_4: PASSED")

    def test_pa_update_invalid_act_val(self):

        self._log.info(
            "\n###################################################\n"
            "TESTCASE_GBP_PA_NEG_5: UPDATE act_value with Invalid "
            "value for act-type=redirect \n"
            "TEST_STEPS::\n"
            "Create a valid Policy Action with 'redirect' action type\n"
            "Update the attributes act_val='invalid value' and name\n"
            "Verify that the update failed and all attrs are having "
            "their original vals\n"
            "###################################################\n")

        act_uuid = self.gbpcfg.gbp_action_config(
            1, 'new_act', action_type='redirect')
        if act_uuid == 0:
            self._log.info("## Step 1: Create Action == Failed")
            return 0
        if self.gbpcfg.gbp_action_config(
                2,
                act_uuid,
                name='grppol_act',
                action_value='INVALID') != 0:
            self._log.info(
                "\n##Step 2: Updating Policy Action's Attrs name & "
                "action_type did NOT Fail")
            return 0
        if self.gbpverify.gbp_action_verify(
                1,
                'new_act',
                id=act_uuid,
                action_type='redirect',
                shared='False') == 0:
            self._log.info(
                "\n## Step 2B: Verify Policy Action Attrs are NOT "
                "updated == Failed")
            return 0
        self._log.info("\n## TESTCASE_GBP_PA_NEG_5: PASSED")

    def test_pa_delete_invalid_pa(self):
        """
        Delete non-existent Policy Action
        """
        self._log.info(
            "\n###################################################\n"
            "TESTCASE_GBP_PA_NEG_6: DELETE NON-EXISTENT/INVALID "
            "POLICY ACTION\n"
            "TEST_STEPS::\n"
            "Delete non-existent Policy Action\n"
            "###################################################\n")
        if self.gbpcfg.gbp_action_config(0, 'grppol_act') != 0:
            self._log.info(
                "\n## Step 1: Expected Error during deletion on "
                "non-existent Policy Action == Failed")
            return 0
        self._log.info("\n## TESTCASE_GBP_PA_NEG_6: PASSED")

if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.contrib.nfp.config_orchestrator.common import (
    topics as a_topics)
from gbpservice.contrib.nfp.config_orchestrator.handlers.config import (
    firewall as fw)
from gbpservice.contrib.nfp.config_orchestrator.handlers.config import (
    loadbalancer as lb)
from gbpservice.contrib.nfp.config_orchestrator.handlers.config import (
    loadbalancerv2 as lbv2)
from gbpservice.contrib.nfp.config_orchestrator.handlers.config import vpn
from gbpservice.contrib.nfp.config_orchestrator.handlers.notification import (
    handler as notif_handler)

from gbpservice.nfp.core.rpc import RpcAgent
from oslo_config import cfg


def rpc_init(sc, conf):
    fwrpcmgr = fw.FwAgent(conf, sc)
    fwagent = RpcAgent(
        sc,
        host=cfg.CONF.host,
        topic=a_topics.FW_NFP_CONFIGAGENT_TOPIC,
        manager=fwrpcmgr
    )

    lb_report_state = {
        'binary': 'NCO',
        'host': cfg.CONF.host,
        'topic': a_topics.LB_NFP_CONFIGAGENT_TOPIC,
        'plugin_topic': a_topics.LB_NFP_PLUGIN_TOPIC,
        'agent_type': 'NFP Loadbalancer agent',
        'configurations': {'device_drivers': ['loadbalancer']},
        'start_flag': True,
        'report_interval': 10
    }
    lbrpcmgr = lb.LbAgent(conf, sc)
    lbagent = RpcAgent(
        sc,
        host=cfg.CONF.host,
        topic=a_topics.LB_NFP_CONFIGAGENT_TOPIC,
        manager=lbrpcmgr,
        report_state=lb_report_state
    )

    lbv2_report_state = {
        'binary': 'NCO',
        'host': cfg.CONF.host,
        'topic': a_topics.LBV2_NFP_CONFIGAGENT_TOPIC,
        'plugin_topic': a_topics.LBV2_NFP_PLUGIN_TOPIC,
        'agent_type': 'NFP Loadbalancer V2 agent',
        'configurations': {'device_drivers': ['loadbalancerv2']},
        'start_flag': True,
        'report_interval': 10
    }
    lbv2rpcmgr = lbv2.Lbv2Agent(conf, sc)
    lbv2agent = RpcAgent(
        sc,
        host=cfg.CONF.host,
        topic=a_topics.LBV2_NFP_CONFIGAGENT_TOPIC,
        manager=lbv2rpcmgr,
        report_state=lbv2_report_state
    )

    vpn_report_state = {
        'binary': 'NCO',
        'host': cfg.CONF.host,
        'topic': a_topics.VPN_NFP_CONFIGAGENT_TOPIC,
        'plugin_topic': a_topics.VPN_NFP_PLUGIN_TOPIC,
        'agent_type': 'NFP Vpn agent',
        'configurations': {'device_drivers': ['vpn']},
        'start_flag': True,
        'report_interval': 10
    }
    vpnrpcmgr = vpn.VpnAgent(conf, sc)
    vpnagent = RpcAgent(
        sc,
        host=cfg.CONF.host,
        topic=a_topics.VPN_NFP_CONFIGAGENT_TOPIC,
        manager=vpnrpcmgr,
        report_state=vpn_report_state
    )

    rpchandler = notif_handler.RpcHandler(conf, sc)
    rpcagent = RpcAgent(
        sc,
        host=cfg.CONF.host,
        topic=a_topics.CONFIG_ORCH_TOPIC,
        manager=rpchandler,
    )

    sc.register_rpc_agents([fwagent, lbagent, lbv2agent, vpnagent, rpcagent])


def nfp_module_init(sc, conf):
    rpc_init(sc, conf)



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""es_ip_allocation_fix_constraints
"""

# revision identifiers, used by Alembic.
revision = '1fadeb573886'
down_revision = '81d13acfbb80'

from alembic import op
from neutron.db import migration
from sqlalchemy.engine import reflection


def upgrade(active_plugins=None, options=None):
    inspector = reflection.Inspector.from_engine(op.get_bind())
    unique_constraints = inspector.get_unique_constraints(
        'gp_es_to_l3p_associations')
    for constraint in unique_constraints:
        if constraint['column_names'] == ['external_segment_id',
                                          'allocated_address']:
            with migration.remove_fks_from_table(
                    'gp_es_to_l3p_associations'):
                op.drop_constraint(constraint['name'],
                                   'gp_es_to_l3p_associations',
                                   'unique')
            break


def downgrade(active_plugins=None, options=None):
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.api.v2 import attributes
from oslo_log import log as logging

from gbpservice.neutron.db.grouppolicy.extensions import group_proxy_db as db
from gbpservice.neutron.db.grouppolicy import group_policy_db as gp_db
from gbpservice.neutron.extensions import driver_proxy_group
from gbpservice.neutron.services.grouppolicy import (
    group_policy_driver_api as api)

LOG = logging.getLogger(__name__)


class ProxyGroupDriver(api.ExtensionDriver):
    _supported_extension_alias = 'proxy_group'
    _extension_dict = driver_proxy_group.EXTENDED_ATTRIBUTES_2_0

    def initialize(self):
        pass

    @property
    def extension_alias(self):
        return self._supported_extension_alias

    @api.default_extension_behavior(db.GroupProxyMapping)
    def process_create_policy_target_group(self, session, data, result):
        data = data['policy_target_group']
        proxied = data.get('proxied_group_id')
        if attributes.is_attr_set(proxied):
            # Set value for proxied group
            record = (session.query(db.GroupProxyMapping).filter_by(
                policy_target_group_id=proxied).first())
            if record:
                if record.proxy_group_id:
                    raise driver_proxy_group.InvalidProxiedGroup(
                        group_id=proxied)
                record.proxy_group_id = result['id']
            else:
                # Record may not exist for that PTG yet
                record = db.GroupProxyMapping(
                    policy_target_group_id=proxied,
                    proxy_group_id=result['id'],
                    proxied_group_id=None)
                session.add(record)
            if not attributes.is_attr_set(data.get('proxy_type')):
                data['proxy_type'] = driver_proxy_group.DEFAULT_PROXY_TYPE
                record = (session.query(db.GroupProxyMapping).filter_by(
                    policy_target_group_id=result['id']).one())
                record.proxy_type = data['proxy_type']
                result['proxy_type'] = data['proxy_type']
            # Proxy PTGs can't have chains enforced
            data['enforce_service_chains'] = False
            record = (session.query(db.GroupProxyMapping).filter_by(
                policy_target_group_id=result['id']).one())
            record.enforce_service_chains = data['enforce_service_chains']
            result['enforce_service_chains'] = data['enforce_service_chains']
        elif attributes.is_attr_set(data.get('proxy_type')):
            raise driver_proxy_group.ProxyTypeSetWithoutProxiedPTG()

    @api.default_extension_behavior(db.GroupProxyMapping)
    def process_update_policy_target_group(self, session, data, result):
        pass

    @api.default_extension_behavior(db.GroupProxyMapping)
    def extend_policy_target_group_dict(self, session, result):
        pass

    @api.default_extension_behavior(db.ProxyGatewayMapping)
    def process_create_policy_target(self, session, data, result):
        self._validate_proxy_gateway(session, data, result)

    @api.default_extension_behavior(db.ProxyGatewayMapping)
    def process_update_policy_target(self, session, data, result):
        self._validate_proxy_gateway(session, data, result)

    @api.default_extension_behavior(db.ProxyGatewayMapping)
    def extend_policy_target_dict(self, session, result):
        pass

    def _validate_proxy_gateway(self, session, data, result):
        data = data['policy_target']
        if data.get('proxy_gateway'):
            ptg_id = result['policy_target_group_id']
            record = session.query(db.GroupProxyMapping).filter_by(
                proxy_group_id=ptg_id).first()
            if not record:
                # The group of this PT is not a proxy
                raise driver_proxy_group.InvalidProxyGatewayGroup(
                    group_id=ptg_id)

    @api.default_extension_behavior(db.ProxyIPPoolMapping)
    def process_create_l3_policy(self, session, data, result):
        data = data['l3_policy']
        gp_db.GroupPolicyDbPlugin.validate_ip_pool(
            data['proxy_ip_pool'], data['ip_version'])
        gp_db.GroupPolicyDbPlugin.validate_subnet_prefix_length(
            data['ip_version'], data['proxy_subnet_prefix_length'],
            data['proxy_ip_pool'])

    @api.default_extension_behavior(db.ProxyIPPoolMapping)
    def process_update_l3_policy(self, session, data, result):
        data = data['l3_policy']
        if 'proxy_subnet_prefix_length' in data:
            gp_db.GroupPolicyDbPlugin.validate_subnet_prefix_length(
                result['ip_version'], data['proxy_subnet_prefix_length'],
                result['proxy_ip_pool'])

    @api.default_extension_behavior(db.ProxyIPPoolMapping)
    def extend_l3_policy_dict(self, session, result):
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.nfp.lib import nfp_context_manager as nfp_ctx_mgr

from gbpservice.nfp.orchestrator.openstack import openstack_driver
from gbpservice.nfp.orchestrator.coal.networking import(
    nfp_network_driver_base as ndb
)


class NFPNeutronNetworkDriver(ndb.NFPNetworkDriverBase):

    def __init__(self, config):
        # self.network_handler = openstack_driver.NeutronClient(config)
        self.neutron_client = openstack_driver.NeutronClient(config)
        self.config = config

    def setup_traffic_steering(self):
        pass

    def create_port(self, token, tenant_id, net_id, name=None):
        with nfp_ctx_mgr.NeutronContextManager as ncm:
            port = ncm.retry(
                self.neutron_client.create_port,
                token, tenant_id, net_id,
                attrs={'name': name})
            return port

    def delete_port(self, token, port_id):
        with nfp_ctx_mgr.NeutronContextManager as ncm:
            ncm.retry(self.neutron_client.delete_port, token, port_id)

    def get_port_id(self, token, port_id):
        return port_id

    def update_port(self, token, port_id, port):
        with nfp_ctx_mgr.NeutronContextManager as ncm:
            port = ncm.retry(self.neutron_client.update_port,
                             token, port_id, **port)
        return port['port']

    def get_port_and_subnet_details(self, token, port_id):
        with nfp_ctx_mgr.NeutronContextManager as ncm:
            port = ncm.retry(self.neutron_client.get_port, token, port_id)

        # ip
        ip = port['port']['fixed_ips'][0]['ip_address']

        # mac
        mac = port['port']['mac_address']

        # gateway ip
        subnet_id = port['port']['fixed_ips'][0]['subnet_id']
        with nfp_ctx_mgr.NeutronContextManager as ncm:
            subnet = ncm.retry(
                self.neutron_client.get_subnet, token, subnet_id)
        cidr = subnet['subnet']['cidr']
        gateway_ip = subnet['subnet']['gateway_ip']

        return (ip, mac, cidr, gateway_ip, port, subnet)

    def get_port_details(self, token, port_id):
        with nfp_ctx_mgr.NeutronContextManager as ncm:
            port = ncm.retry(self.neutron_client.get_port, token, port_id)

        # ip
        ip = port['port']['fixed_ips'][0]['ip_address']

        # mac
        mac = port['port']['mac_address']

        # gateway ip
        subnet_id = port['port']['fixed_ips'][0]['subnet_id']
        with nfp_ctx_mgr.NeutronContextManager as ncm:
            subnet = ncm.retry(
                self.neutron_client.get_subnet, token, subnet_id)
        cidr = subnet['subnet']['cidr']
        gateway_ip = subnet['subnet']['gateway_ip']

        return (ip, mac, cidr, gateway_ip, port, subnet)

    def set_promiscuos_mode(self, token, port_id, enable_port_security):
        port_security = bool(enable_port_security)
        with nfp_ctx_mgr.NeutronContextManager as ncm:
            ncm.retry(self.neutron_client.update_port, token, port_id,
                      security_groups=[],
                      port_security_enabled=port_security)

    def get_service_profile(self, token, service_profile_id):
        return {}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from neutron.callbacks import events
from neutron.callbacks import registry
from neutron.callbacks import resources
from neutron.common import constants as l3_constants
from neutron.db import api as db_api
from neutron.db import l3_db
from neutron.db import models_v2
from neutron.db import securitygroups_db
from neutron.plugins.ml2 import db as ml2_db
from oslo_log import log
from sqlalchemy import event

LOG = log.getLogger(__name__)


# Monkey patch create floatingip to allow subnet_id to be specified.
# this will only be valid for internal calls, and can't be exploited from the
# API.
def create_floatingip(
        self, context, floatingip,
        initial_status=l3_db.l3_constants.FLOATINGIP_STATUS_ACTIVE):
    fip = floatingip['floatingip']
    tenant_id = self._get_tenant_id_for_create(context, fip)
    fip_id = l3_db.uuidutils.generate_uuid()

    f_net_id = fip['floating_network_id']
    if not self._core_plugin._network_is_external(context, f_net_id):
        msg = _("Network %s is not a valid external network") % f_net_id
        raise l3_db.n_exc.BadRequest(resource='floatingip', msg=msg)

    with context.session.begin(subtransactions=True):
        # This external port is never exposed to the tenant.
        # it is used purely for internal system and admin use when
        # managing floating IPs.

        port = {'tenant_id': '',  # tenant intentionally not set
                'network_id': f_net_id,
                'mac_address': l3_db.attributes.ATTR_NOT_SPECIFIED,
                'fixed_ips': l3_db.attributes.ATTR_NOT_SPECIFIED,
                'admin_state_up': True,
                'device_id': fip_id,
                'device_owner': l3_db.DEVICE_OWNER_FLOATINGIP,
                'status': l3_db.l3_constants.PORT_STATUS_NOTAPPLICABLE,
                'name': ''}

        if fip.get('floating_ip_address'):
            port['fixed_ips'] = [{'ip_address': fip['floating_ip_address']}]
        if fip.get('subnet_id'):
            port['fixed_ips'] = [{'subnet_id': fip['subnet_id']}]

        external_port = self._core_plugin.create_port(context.elevated(),
                                                      {'port': port})

        # Ensure IP addresses are allocated on external port
        if not external_port['fixed_ips']:
            raise l3_db.n_exc.ExternalIpAddressExhausted(net_id=f_net_id)

        floating_fixed_ip = external_port['fixed_ips'][0]
        floating_ip_address = floating_fixed_ip['ip_address']
        floatingip_db = l3_db.FloatingIP(
            id=fip_id,
            tenant_id=tenant_id,
            status=initial_status,
            floating_network_id=fip['floating_network_id'],
            floating_ip_address=floating_ip_address,
            floating_port_id=external_port['id'])
        fip['tenant_id'] = tenant_id
        # Update association with internal port
        # and define external IP address
        self._update_fip_assoc(context, fip,
                               floatingip_db, external_port)
        context.session.add(floatingip_db)

    return self._make_floatingip_dict(floatingip_db)

l3_db.L3_NAT_dbonly_mixin.create_floatingip = create_floatingip


# REVISIT(ivar): Monkey patch to allow explicit router_id to be set in Neutron
# for Floating Ip creation (for internal calls only). Once we split the server,
# this could be part of a GBP Neutron L3 driver.
def _get_assoc_data(self, context, fip, floating_network_id):
    (internal_port, internal_subnet_id,
     internal_ip_address) = self._internal_fip_assoc_data(context, fip)
    if fip.get('router_id'):
        router_id = fip['router_id']
        del fip['router_id']
    else:
        router_id = self._get_router_for_floatingip(context,
                                                    internal_port,
                                                    internal_subnet_id,
                                                    floating_network_id)

    return fip['port_id'], internal_ip_address, router_id


def _update_fip_assoc(self, context, fip, floatingip_db, external_port):
    previous_router_id = floatingip_db.router_id
    port_id, internal_ip_address, router_id = (
        self._check_and_get_fip_assoc(context, fip, floatingip_db))
    update = {'fixed_ip_address': internal_ip_address,
              'fixed_port_id': port_id,
              'router_id': router_id,
              'last_known_router_id': previous_router_id}
    if 'description' in fip:
        update['description'] = fip['description']
    floatingip_db.update(update)
    next_hop = None
    if router_id:
        router = self._get_router(context.elevated(), router_id)
        gw_port = router.gw_port
        if gw_port:
            for fixed_ip in gw_port.fixed_ips:
                addr = netaddr.IPAddress(fixed_ip.ip_address)
                if addr.version == l3_constants.IP_VERSION_4:
                    next_hop = fixed_ip.ip_address
                    break
    args = {'fixed_ip_address': internal_ip_address,
            'fixed_port_id': port_id,
            'router_id': router_id,
            'last_known_router_id': previous_router_id,
            'floating_ip_address': floatingip_db.floating_ip_address,
            'floating_network_id': floatingip_db.floating_network_id,
            'next_hop': next_hop,
            'context': context}
    registry.notify(resources.FLOATING_IP,
                    events.AFTER_UPDATE,
                    self._update_fip_assoc,
                    **args)

l3_db.L3_NAT_dbonly_mixin._get_assoc_data = _get_assoc_data
l3_db.L3_NAT_dbonly_mixin._update_fip_assoc = _update_fip_assoc


# REVISIT(ivar): Neutron adds a tenant filter on SG lookup for a given port,
# this breaks our service chain plumbing model so for now we should monkey
# patch the specific method. A follow up with the Neutron team is needed to
# figure out the reason for this and how to proceed for future releases.
def _get_security_groups_on_port(self, context, port):
    """Check that all security groups on port belong to tenant.

    :returns: all security groups IDs on port belonging to tenant.
    """
    p = port['port']
    if not securitygroups_db.attributes.is_attr_set(
            p.get(securitygroups_db.ext_sg.SECURITYGROUPS)):
        return
    if p.get('device_owner') and p['device_owner'].startswith('network:'):
        return

    port_sg = p.get(securitygroups_db.ext_sg.SECURITYGROUPS, [])
    filters = {'id': port_sg}
    valid_groups = set(g['id'] for g in
                       self.get_security_groups(context, fields=['id'],
                                                filters=filters))

    requested_groups = set(port_sg)
    port_sg_missing = requested_groups - valid_groups
    if port_sg_missing:
        raise securitygroups_db.ext_sg.SecurityGroupNotFound(
            id=', '.join(port_sg_missing))

    return requested_groups

securitygroups_db.SecurityGroupDbMixin._get_security_groups_on_port = (
    _get_security_groups_on_port)


def get_port_from_device_mac(context, device_mac):
    LOG.debug("get_port_from_device_mac() called for mac %s", device_mac)
    qry = context.session.query(models_v2.Port).filter_by(
        mac_address=device_mac).order_by(models_v2.Port.device_owner.desc())
    return qry.first()

ml2_db.get_port_from_device_mac = get_port_from_device_mac

PUSH_NOTIFICATIONS_METHOD = None
DISCARD_NOTIFICATIONS_METHOD = None


def get_session(autocommit=True, expire_on_commit=False, use_slave=False):
    # The folowing are declared as global so that they can
    # used in the inner functions that follow.
    global PUSH_NOTIFICATIONS_METHOD
    global DISCARD_NOTIFICATIONS_METHOD
    from gbpservice.network.neutronv2 import local_api
    PUSH_NOTIFICATIONS_METHOD = (
        local_api.post_notifications_from_queue)
    DISCARD_NOTIFICATIONS_METHOD = (
        local_api.discard_notifications_after_rollback)

    # The following two lines are copied from the original
    # implementation of db_api.get_session() and should be updated
    # if the original implementation changes.
    facade = db_api._create_facade_lazily()
    new_session = facade.get_session(autocommit=autocommit,
                                     expire_on_commit=expire_on_commit,
                                     use_slave=use_slave)

    new_session.notification_queue = {}

    def gbp_after_transaction(session, transaction):
        if transaction and not transaction._parent and (
            not transaction.is_active and not transaction.nested):
            if transaction in session.notification_queue:
                # push the queued notifications only when the
                # outermost transaction completes
                PUSH_NOTIFICATIONS_METHOD(session, transaction)

    def gbp_after_rollback(session):
        # We discard all queued notifiactions if the transaction fails.
        DISCARD_NOTIFICATIONS_METHOD(session)

    if local_api.BATCH_NOTIFICATIONS:
        event.listen(new_session, "after_transaction_end",
                     gbp_after_transaction)
        event.listen(new_session, "after_rollback",
                     gbp_after_rollback)

    return new_session


db_api.get_session = get_session



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""intra_ptg_allow

Revision ID: fce38a8588a2
Revises: 4af01d620224
Create Date: 2016-12-06 17:01:30.735865

"""

# revision identifiers, used by Alembic.
revision = 'fce38a8588a2'
down_revision = '4af01d620224'

from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'gp_apic_intra_ptg',
        sa.Column('policy_target_group_id', sa.String(length=36),
                  nullable=False),
        sa.Column('intra_ptg_allow', sa.Boolean,
                  server_default=sa.sql.true(), nullable=False),
        sa.ForeignKeyConstraint(
            ['policy_target_group_id'], ['gp_policy_target_groups.id'],
            name='gp_apic_intra_ptg_fk_ptgid', ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('policy_target_group_id')
    )


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.tests import base

from gbpservice.contrib.nfp.configurator.drivers.nfp_service.heat.heat_driver \
                                                        import HeatDriver
from gbpservice.contrib.nfp.configurator.lib import (
                                            nfp_service_constants as const)
from gbpservice.contrib.tests.unit.nfp.configurator.test_data import (
                                                nfp_service_test_data as fo)


class NfpServiceHeatDriverTestCase(base.BaseTestCase):
    """ Implements test cases for driver methods
    of nfp service.

    """

    def __init__(self, *args, **kwargs):
        super(NfpServiceHeatDriverTestCase, self).__init__(*args, **kwargs)
        self.fo = fo.FakeObjects()

    def test_configure_interfaces(self):
        """ Implements test case for nfp service heat driver.

        Returns: none

        """

        driver = HeatDriver(self.fo.conf)
        actual_val = driver.run_heat(self.fo.context, self.fo.kwargs)

        expected_val = const.UNHANDLED_RESULT
        self.assertEqual(actual_val, expected_val)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import fixtures
from neutron import context
from neutron.db import api as db_api
from neutron.tests import base

from gbpservice.nfp.common import constants as nfp_constants
from gbpservice.nfp.common import exceptions as nfp_exc
from gbpservice.nfp.orchestrator.db import nfp_db
from gbpservice.nfp.orchestrator.db import nfp_db_model
from sqlalchemy.orm import exc
import uuid


class SqlFixture(fixtures.Fixture):

    # flag to indicate that the models have been loaded
    _TABLES_ESTABLISHED = False

    def _setUp(self):
        # Register all data models
        engine = db_api.get_engine()
        if not SqlFixture._TABLES_ESTABLISHED:
            nfp_db_model.BASE.metadata.create_all(engine)
            SqlFixture._TABLES_ESTABLISHED = True

        def clear_tables():
            with engine.begin() as conn:
                for table in reversed(
                        nfp_db_model.BASE.metadata.sorted_tables):
                    conn.execute(table.delete())

        self.addCleanup(clear_tables)


class SqlTestCaseLight(base.DietTestCase):
    """All SQL taste, zero plugin/rpc sugar"""

    def setUp(self):
        super(SqlTestCaseLight, self).setUp()
        self.useFixture(SqlFixture())


class SqlTestCase(base.BaseTestCase):

    def setUp(self):
        super(SqlTestCase, self).setUp()
        self.useFixture(SqlFixture())


class NFPDB(nfp_db.NFPDbBase):
    pass


class NFPDBTestCase(SqlTestCase):

    def setUp(self):
        super(NFPDBTestCase, self).setUp()
        self.ctx = context.get_admin_context()
        self.nfp_db = NFPDB()
        self.session = db_api.get_session()

    def create_network_function(self, attributes=None):
        if attributes is None:
            attributes = {
                'name': 'name',
                'description': 'description',
                'tenant_id': 'tenant_id',
                'service_id': 'service_id',
                'service_chain_id': 'service_chain_id',
                'service_profile_id': 'service_profile_id',
                'service_config': 'service_config',
                'config_policy_id': 'config_policy_id',
                'status': 'status'
            }
        return self.nfp_db.create_network_function(self.session, attributes)

    def test_create_network_function(self):
        attrs = {
            'name': 'name',
            'description': 'description',
            'tenant_id': 'tenant_id',
            'service_id': 'service_id',
            'service_chain_id': 'service_chain_id',
            'service_profile_id': 'service_profile_id',
            'service_config': 'service_config',
            'config_policy_id': 'config_policy_id',
            'status': 'status'
        }

        network_function = self.create_network_function(attrs)
        for key in attrs:
            self.assertEqual(attrs[key], network_function[key])
        self.assertIsNotNone(network_function['id'])

    def test_create_network_function_with_mandatory_values(self):
        attrs_mandatory = {
            'name': 'name',
            'tenant_id': 'tenant_id',
            'service_id': 'service_id',
            'service_profile_id': 'service_profile_id',
            'status': 'status'
        }
        network_function = self.create_network_function(attrs_mandatory)
        for key in attrs_mandatory:
            self.assertEqual(attrs_mandatory[key], network_function[key])
        self.assertIsNotNone(network_function['id'])
        non_mandatory_args = ['service_chain_id', 'service_config',
                              'config_policy_id']
        for arg in non_mandatory_args:
            self.assertIsNone(network_function[arg])

    def test_get_network_function(self):
        attrs_all = {
            'name': 'name',
            'description': 'description',
            'tenant_id': 'tenant_id',
            'service_id': 'service_id',
            'service_chain_id': 'service_chain_id',
            'service_profile_id': 'service_profile_id',
            'service_config': 'service_config',
            'config_policy_id': 'config_policy_id',
            'status': 'status'
        }
        network_function = self.create_network_function(attrs_all)
        db_network_function = self.nfp_db.get_network_function(
            self.session, network_function['id'])
        for key in attrs_all:
            self.assertEqual(attrs_all[key], db_network_function[key])

    def test_list_network_function(self):
        network_function = self.create_network_function()
        network_functions = self.nfp_db.get_network_functions(self.session)
        self.assertEqual(1, len(network_functions))
        self.assertEqual(network_function['id'], network_functions[0]['id'])

    def test_list_network_function_with_filters(self):
        attrs = {
            'name': 'name',
            'tenant_id': 'tenant_id',
            'service_id': 'service_id',
            'service_profile_id': 'service_profile_id',
            'status': 'status'
        }
        network_function = self.create_network_function(attrs)
        filters = {'service_id': ['service_id']}
        network_functions = self.nfp_db.get_network_functions(
            self.session, filters=filters)
        self.assertEqual(1, len(network_functions))
        self.assertEqual(network_function['id'], network_functions[0]['id'])
        filters = {'service_id': ['nonexisting']}
        network_functions = self.nfp_db.get_network_functions(
            self.session, filters=filters)
        self.assertEqual([], network_functions)

    def test_update_network_function(self):
        network_function = self.create_network_function()
        self.assertIsNotNone(network_function['id'])
        updated_network_function = {'status': 'ERROR'}
        network_function = self.nfp_db.update_network_function(
            self.session, network_function['id'], updated_network_function)
        self.assertEqual('ERROR', network_function['status'])

    def test_delete_network_function(self):
        network_function = self.create_network_function()
        self.assertIsNotNone(network_function['id'])
        self.nfp_db.delete_network_function(
            self.session, network_function['id'])
        self.assertRaises(nfp_exc.NetworkFunctionNotFound,
                          self.nfp_db.get_network_function,
                          self.session, network_function['id'])

    def create_network_function_instance(self, attributes=None,
                                         create_nfd=True):
        if attributes is None:
            nfd = (self.create_network_function_device()['id']
                   if create_nfd else None)
            attributes = {
                'name': 'name',
                'description': 'description',
                'tenant_id': 'tenant_id',
                'network_function_id': self.create_network_function()['id'],
                'network_function_device_id': nfd,
                'ha_state': "Active",
                'port_info': [
                    {'id': 'myportid1',
                     'port_model': nfp_constants.NEUTRON_PORT,
                     'port_classification': nfp_constants.PROVIDER,
                     'port_role': nfp_constants.ACTIVE_PORT},
                    {'id': 'myportid2',
                     'port_model': nfp_constants.GBP_PORT,
                     'port_classification': nfp_constants.CONSUMER,
                     'port_role': nfp_constants.MASTER_PORT}
                ],
                'status': 'status'
            }
        return self.nfp_db.create_network_function_instance(
            self.session, attributes)

    def test_create_network_function_instance(self):
        network_function = self.create_network_function()
        attrs = {
            'name': 'name',
            'description': 'description',
            'tenant_id': 'tenant_id',
            'network_function_id': network_function['id'],
            'network_function_device_id': (
                self.create_network_function_device()['id']),
            'ha_state': 'Active',
            'port_info': [
                {'id': 'my_nfi_port_id1',
                 'port_model': nfp_constants.NEUTRON_PORT,
                 'port_classification': nfp_constants.PROVIDER,
                 'port_role': nfp_constants.ACTIVE_PORT},
                {'id': 'my_nfi_port_id2',
                 'port_model': nfp_constants.GBP_PORT,
                 'port_classification': nfp_constants.CONSUMER,
                 'port_role': nfp_constants.MASTER_PORT}
            ],
            'status': 'status'
        }
        network_function_instance = (
            self.nfp_db.create_network_function_instance(self.session, attrs))
        for key in attrs:
            self.assertEqual(attrs[key], network_function_instance[key])
        self.assertIsNotNone(network_function_instance['id'])

    def test_create_network_function_instance_mandatory_values(self):
        network_function = self.create_network_function()
        attrs_mandatory = {
            'name': 'name',
            'tenant_id': 'tenant_id',
            'network_function_id': network_function['id'],
            'status': 'status',
            'port_info': []
        }
        network_function_instance = (
            self.nfp_db.create_network_function_instance(
                self.session, attrs_mandatory))
        for key in attrs_mandatory:
            self.assertEqual(attrs_mandatory[key],
                             network_function_instance[key])
        self.assertIsNotNone(network_function_instance['id'])
        non_mandatory_args = ['network_function_device_id', 'ha_state']
        for arg in non_mandatory_args:
            self.assertIsNone(network_function_instance[arg])
        self.assertEqual([], network_function_instance['port_info'])

    def test_get_network_function_instance(self):
        network_function = self.create_network_function()
        attrs_all = {
            'name': 'name',
            'description': 'description',
            'tenant_id': 'tenant_id',
            'network_function_id': network_function['id'],
            'network_function_device_id': (
                self.create_network_function_device()['id']),
            'ha_state': 'Active',
            'port_info': [
                {'id': 'my_nfi_port_id1',
                 'port_model': nfp_constants.NEUTRON_PORT,
                 'port_classification': nfp_constants.PROVIDER,
                 'port_role': nfp_constants.ACTIVE_PORT},
                {'id': 'my_nfi_port_id2',
                 'port_model': nfp_constants.GBP_PORT,
                 'port_classification': nfp_constants.CONSUMER,
                 'port_role': nfp_constants.MASTER_PORT}
            ],
            'status': 'status'
        }
        network_function_instance = (
            self.nfp_db.create_network_function_instance(
                self.session, attrs_all))
        db_network_function_instance = (
            self.nfp_db.get_network_function_instance(
                self.session, network_function_instance['id']))
        for key in attrs_all:
            self.assertEqual(attrs_all[key], db_network_function_instance[key])

    def test_list_network_function_instance(self):
        self.test_create_network_function_instance()
        nf_instances = self.nfp_db.get_network_function_instances(
            self.session)
        self.assertEqual(1, len(nf_instances))

    def test_list_network_function_instances_with_filters(self):
        self.test_create_network_function_instance()
        filters = {'ha_state': ['Active']}
        nf_instances = self.nfp_db.get_network_function_instances(
            self.session, filters=filters)
        self.assertEqual(1, len(nf_instances))
        filters = {'ha_state': ['nonexisting']}
        nf_instances = self.nfp_db.get_network_function_instances(
            self.session, filters=filters)
        self.assertEqual([], nf_instances)

    def test_update_network_function_instance(self):
        network_function_instance = self.create_network_function_instance()
        self.assertIsNotNone(network_function_instance['id'])
        updated_nfi = {'status': 'ERROR'}
        nf_instance = self.nfp_db.update_network_function_instance(
            self.session, network_function_instance['id'], updated_nfi)
        self.assertEqual('ERROR', nf_instance['status'])

    def test_delete_network_function_instance(self):
        network_function_instance = self.create_network_function_instance()
        port_info = network_function_instance['port_info']
        self.assertIsNotNone(network_function_instance['id'])
        self.nfp_db.delete_network_function_instance(
            self.session, network_function_instance['id'])
        self.assertRaises(nfp_exc.NetworkFunctionInstanceNotFound,
                          self.nfp_db.get_network_function_instance,
                          self.session, network_function_instance['id'])
        for port_id in port_info:
            self.assertRaises(nfp_exc.NFPPortNotFound,
                              self.nfp_db.get_port_info,
                              self.session,
                              port_id)

    def create_network_function_device(self, attributes=None):
        if attributes is None:
            attributes = {
                'name': 'name',
                'description': 'description',
                'tenant_id': 'tenant_id',
                'mgmt_ip_address': 'mgmt_ip_address',
                'monitoring_port_id': {
                    'id': 'myid1_ha_port',
                    'port_model': nfp_constants.NEUTRON_PORT,
                    'port_classification': nfp_constants.MONITOR,
                    'port_role': nfp_constants.ACTIVE_PORT
                },
                'monitoring_port_network': {
                    'id': 'mynetwork_id',
                    'network_model': nfp_constants.NEUTRON_NETWORK
                },
                'service_vendor': 'service_vendor',
                'max_interfaces': 3,
                'reference_count': 2,
                'interfaces_in_use': 1,
                'mgmt_port_id': {
                    'id': 'myid1',
                    'port_model': nfp_constants.NEUTRON_PORT,
                    'port_classification': nfp_constants.MANAGEMENT,
                    'port_role': nfp_constants.ACTIVE_PORT},
                'status': 'status'
            }
        return self.nfp_db.create_network_function_device(
            self.session, attributes)

    def test_create_network_function_device(self):
        attrs = {
            'name': 'name',
            'description': 'description',
            'tenant_id': 'tenant_id',
            'mgmt_ip_address': 'mgmt_ip_address',
            'monitoring_port_id': {
                'id': 'myid1_ha_port',
                'port_model': nfp_constants.NEUTRON_PORT,
                'port_classification': nfp_constants.MONITOR,
                'port_role': nfp_constants.ACTIVE_PORT
            },
            'monitoring_port_network': {
                'id': 'mynetwork_id',
                'network_model': nfp_constants.NEUTRON_NETWORK
            },
            'service_vendor': 'service_vendor',
            'max_interfaces': 3,
            'reference_count': 2,
            'interfaces_in_use': 1,
            'mgmt_port_id': {
                'id': 'myid1',
                'port_model': nfp_constants.NEUTRON_PORT,
                'port_classification': nfp_constants.MANAGEMENT,
                'port_role': nfp_constants.ACTIVE_PORT},
            'status': 'status'
        }
        network_function_device = self.nfp_db.create_network_function_device(
            self.session, attrs)
        self.assertIn('gateway_port', network_function_device)
        for key in attrs:
            if (key == 'mgmt_port_id') or (key == 'monitoring_port_id'):
                self.assertEqual(attrs[key]['id'],
                                 network_function_device[key])
                continue
            self.assertEqual(attrs[key], network_function_device[key])
        self.assertIsNotNone(network_function_device['id'])

    def test_create_network_function_device_mandatory_values(self):
        attrs_mandatory = {
            'name': 'name',
            'tenant_id': 'tenant_id',
            'mgmt_ip_address': 'mgmt_ip_address',
            'service_vendor': 'service_vendor',
            'max_interfaces': 3,
            'reference_count': 2,
            'interfaces_in_use': 1,
            'status': 'status'
        }
        nf_device = self.nfp_db.create_network_function_device(
            self.session, attrs_mandatory)
        for key in attrs_mandatory:
            self.assertEqual(attrs_mandatory[key], nf_device[key])
        self.assertIsNotNone(nf_device['id'])
        non_mandatory_args = ['monitoring_port_id',
                              'monitoring_port_network']
        for arg in non_mandatory_args:
            self.assertIsNone(nf_device[arg])
        self.assertEqual(None, nf_device['mgmt_port_id'])

    def test_get_network_function_device(self):
        attrs = {
            'name': 'name',
            'description': 'description',
            'tenant_id': 'tenant_id',
            'mgmt_ip_address': 'mgmt_ip_address',
            'monitoring_port_id': {
                'id': 'myid1_ha_port',
                'port_model': nfp_constants.NEUTRON_PORT,
                'port_classification': nfp_constants.MONITOR,
                'port_role': nfp_constants.ACTIVE_PORT
            },
            'monitoring_port_network': {
                'id': 'mynetwork_id',
                'network_model': nfp_constants.NEUTRON_NETWORK
            },
            'service_vendor': 'service_vendor',
            'max_interfaces': 3,
            'reference_count': 2,
            'interfaces_in_use': 1,
            'mgmt_port_id': {
                'id': 'myid1',
                'port_model': nfp_constants.NEUTRON_PORT,
                'port_classification': nfp_constants.MANAGEMENT,
                'port_role': nfp_constants.ACTIVE_PORT},
            'status': 'status'
        }
        network_function_device = self.nfp_db.create_network_function_device(
            self.session, attrs)
        db_network_function_device = self.nfp_db.get_network_function_device(
            self.session, network_function_device['id'])
        for key in attrs:
            if (key == 'mgmt_port_id') or (key == 'monitoring_port_id'):
                self.assertEqual(attrs[key]['id'],
                                 network_function_device[key])
                continue
            self.assertEqual(attrs[key], db_network_function_device[key])

    def test_list_network_function_device(self):
        self.test_create_network_function_device()
        network_function_devices = self.nfp_db.get_network_function_devices(
            self.session)
        self.assertEqual(1, len(network_function_devices))

    def test_list_network_function_devices_with_filters(self):
        self.test_create_network_function_device()
        filters = {'service_vendor': ['service_vendor']}
        network_function_devices = self.nfp_db.get_network_function_devices(
            self.session, filters=filters)
        self.assertEqual(1, len(network_function_devices))
        filters = {'service_vendor': ['nonexisting']}
        network_function_devices = self.nfp_db.get_network_function_devices(
            self.session, filters=filters)
        self.assertEqual([], network_function_devices)

    def test_update_network_function_device(self):
        attrs = {
            'name': 'name',
            'description': 'description',
            'tenant_id': 'tenant_id',
            'mgmt_ip_address': 'mgmt_ip_address',
            'monitoring_port_id': {
                'id': 'myid1_ha_port',
                'port_model': nfp_constants.NEUTRON_PORT,
                'port_classification': nfp_constants.MONITOR,
                'port_role': nfp_constants.ACTIVE_PORT
            },
            'monitoring_port_network': {
                'id': 'mynetwork_id',
                'network_model': nfp_constants.NEUTRON_NETWORK
            },
            'service_vendor': 'service_vendor',
            'max_interfaces': 3,
            'reference_count': 2,
            'interfaces_in_use': 1,
            'mgmt_port_id': {
                'id': 'myid1',
                'port_model': nfp_constants.NEUTRON_PORT,
                'port_classification': nfp_constants.MANAGEMENT,
                'port_role': nfp_constants.ACTIVE_PORT},
            'status': 'status'
        }
        network_function_device = self.nfp_db.create_network_function_device(
            self.session, attrs)
        for key in attrs:
            if (key == 'mgmt_port_id') or (key == 'monitoring_port_id'):
                self.assertEqual(attrs[key]['id'],
                                 network_function_device[key])
                continue

            self.assertEqual(attrs[key], network_function_device[key])
        self.assertIsNotNone(network_function_device['id'])

        # update name
        updated_network_function_device = {
            'name': 'new_name'
        }
        updated_nfd = self.nfp_db.update_network_function_device(
            self.session,
            network_function_device['id'],
            updated_network_function_device)
        self.assertEqual('new_name', updated_nfd['name'])
        del updated_nfd['name']
        for key in attrs:
            if (key == 'mgmt_port_id') or (key == 'monitoring_port_id'):
                self.assertEqual(attrs[key]['id'],
                                 network_function_device[key])
                continue
            if key != 'name':
                self.assertEqual(attrs[key], updated_nfd[key])

        # Update mgmt port
        updated_network_function_device = {
            'mgmt_port_id': {
                'id': 'myid3',
                'port_model': nfp_constants.NEUTRON_PORT,
                'port_classification': nfp_constants.MANAGEMENT,
                'port_role': nfp_constants.ACTIVE_PORT},
            'name': 'name'
        }
        updated_nfd = self.nfp_db.update_network_function_device(
            self.session,
            network_function_device['id'],
            copy.deepcopy(updated_network_function_device))
        self.assertEqual(updated_nfd['mgmt_port_id'], 'myid3')
        del updated_nfd['mgmt_port_id']
        for key in attrs:
            if (key != 'mgmt_port_id') and (key != 'monitoring_port_id'):
                self.assertEqual(attrs[key], updated_nfd[key])

    def test_delete_network_function_device(self):
        network_function_device = self.create_network_function_device()
        mgmt_port_id = network_function_device['mgmt_port_id']
        self.assertIsNotNone(network_function_device['id'])
        self.nfp_db.delete_network_function_device(
            self.session, network_function_device['id'])
        self.assertRaises(nfp_exc.NetworkFunctionDeviceNotFound,
                          self.nfp_db.get_network_function_device,
                          self.session, network_function_device['id'])
        self.assertRaises(nfp_exc.NFPPortNotFound,
                          self.nfp_db.get_port_info,
                          self.session,
                          mgmt_port_id)

    def _get_gateway_details(self):
        return dict(
                id=str(uuid.uuid4()),
                network_function_id=self.create_network_function()['id'],
                gw_ptg=str(uuid.uuid4()),
                primary_instance_gw_pt=str(uuid.uuid4()),
                secondary_instance_gw_pt=str(uuid.uuid4()),
                gateway_vips=[dict(id=str(uuid.uuid4()))]
        )

    def test_add_service_gateway_details(self):
        gateway_details = self._get_gateway_details()
        gateway = self.nfp_db.add_service_gateway_details(
                self.session, gateway_details)
        self.assertIsNotNone(gateway['id'])
        self.nfp_db.delete_network_function(
                self.session, gateway_details['network_function_id'])
        gateway_details.update(
                network_function_id=self.create_network_function()['id'],
                gateway_vips=dict(primary_gw_vip_pt=str(uuid.uuid4()),
                                  secondary_gw_vip_pt=str(uuid.uuid4())))
        gateway = self.nfp_db.add_service_gateway_details(
                self.session, gateway_details)
        self.assertIsNotNone(gateway['id'])

    def test_get_gateway_detail(self):
        gateway_details = self._get_gateway_details()
        gateway = self.nfp_db.add_service_gateway_details(
                self.session, gateway_details)
        self.assertIsNotNone(gateway['id'])
        _gateway = self.nfp_db.get_gateway_detail(
                self.session, gateway_details['network_function_id'])
        self.assertEqual((gateway['id'], gateway['network_function_id']),
                         (_gateway['id'], _gateway['network_function_id']))

    def test_get_providers_for_gateway(self):
        gateway_details = self._get_gateway_details()
        gateway = self.nfp_db.add_service_gateway_details(
                self.session, gateway_details)
        self.assertIsNotNone(gateway['id'])
        _gateway = self.nfp_db.get_providers_for_gateway(
                self.session, gateway_details['gw_ptg'])[0]
        self.assertEqual((gateway['id'], gateway['network_function_id']),
                         (_gateway['id'], _gateway['network_function_id']))

    def test_delete_gateway(self):
        gateway_details = self._get_gateway_details()
        gateway = self.nfp_db.add_service_gateway_details(
                self.session, gateway_details)
        self.assertIsNotNone(gateway['id'])
        self.nfp_db.delete_network_function(self.session, gateway_details[
            'network_function_id'])
        self.assertRaises(exc.NoResultFound, self.nfp_db.get_gateway_detail,
            self.session, gateway_details['network_function_id'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.nfp.core import log as nfp_logging


from gbpservice.contrib.nfp.configurator.drivers.base import base_driver
from gbpservice.contrib.nfp.configurator.lib import (
    nfp_service_constants as const)

LOG = nfp_logging.getLogger(__name__)


class HeatDriver(base_driver.BaseDriver):
    """ Heat as a driver for handling config script
        heat configuration requests.

    We initialize service type in this class because agent loads
    class object only for those driver classes that have service type
    initialized. Also, only this driver class is exposed to the agent.

    """

    service_type = const.SERVICE_TYPE
    resource_type = const.HEAT_RESOURCE

    def __init__(self, conf):
        pass

    def run_heat(self, context, kwargs):
        msg = ("Heat template execution request received but unhandled")
        LOG.info(msg)
        return const.UNHANDLED_RESULT



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from gbpservice.contrib.nfp.configurator.lib import (
    filter_constants as constants)


class Filter(object):
    """ Filter class which provides data asked in a specific format.

    This class mocks all rpc calls going from *aaS agent/driver to respective
        *aaS plugin.
    """

    def __init__(self, topic, default_version):
        pass

    def call(self, context, msg):
        """Returns data in specific format after applying filter on context

        :param context
        :param msg e.g  {'args': {'key': value,..},'method': 'function_name'}}

        Returns: data after applying filter on it

        """
        filters = {}
        try:
            for fk, fv in msg['args'].items():
                if dict == type(fv):
                    filters = fv
                    break
                if fv:
                    filters = {fk: fv}
                    break

            method = getattr(self, '_%s' % (msg['method']))
            return method(context, filters)
        except Exception as e:
            raise e

    def make_msg(self, method, **kwargs):
        """ Helper function needed to invoke Filter.call()
        :param method - method name
        :kwargs kwargs - filters to be used

        Returns: dict
        """
        return {'method': method,
                'args': kwargs}

    def apply_filter(self, data, filters):
        """ Apply filters on data

        :param filters e.g  {k:[v],k:[v]}
        :param data e.g [{k:v,k:v,k:v},
                      {k:v,k:v,k:v},
                      {k:v,k:v}]

        Returns: data after applying filter on it

        """

        for fk, fv in filters.items():
            for d in data[:]:
                if d.get(fk) is None:
                    data.remove(d)
                if fk in d and d[fk] != fv[0]:
                    data.remove(d)
        return data

    def get_record(self, data, key, value):
        """Get single record based on key and value

        :param data
        :praam key
        :param value

        Returns: record
        """
        for d in data:
            if key in d and d[key] == value:
                return d

    def _get_vpn_services(self, context, filters):
        """ Get vpn service from context after applying filter

        :param context - vpn related resources
        e.g context = {'service_info':{'vpnservices': [vpnservices],
                                       'ikepolicies': [ikepolicies],
                                       'ipsecpolicies':[ipsecpolicies],
                                       'ipsec_site_conns':
                                                [ipsec_site_connections],
                                        'routers': [routers],
                                        'subnets': [subnets]
                                      }
                      }
        :param filters
        e.g { 'ids' : [vpn service ids],
              'filters': filters }

        Returns: [vpn services]

        """
        vpn_ids = None
        if 'ids' in filters and filters['ids']:
            vpn_ids = filters['ids']
        service_info = context['service_info']
        vpnservices = service_info['vpnservices']
        filtered_vpns = []
        if vpn_ids:
            for vpn_id in vpn_ids:
                filtered_vpns.append(
                    self.get_record(vpnservices, 'id', vpn_id))
            return filtered_vpns
        else:
            return self.apply_filter(vpnservices, filters)

    def _get_ipsec_conns(self, context, filters):
        """ Get ipsec site conns from context after applying filter

        :param context - vpn related resources
        e.g context = {'service_info':{'vpnservices': [vpnservices],
                                       'ikepolicies': [ikepolicies],
                                       'ipsecpolicies':[ipsecpolicies],
                                       'ipsec_site_conns':
                                                [ipsec_site_connections],
                                        'routers': [routers],
                                        'subnets': [subnets]
                                      }
                      }
        :param filters e.g { 'tenant_id': [tenant_id],
                             'peer_address': [conn['peer_address']]
                           }

        Returns: [ipsec site conns]

        """
        service_info = context['service_info']
        ipsec_conns = copy.deepcopy(service_info['ipsec_site_conns'])

        return self.apply_filter(ipsec_conns, filters)

    def _get_vpn_servicecontext(self, context, filters):
        """Get vpnservice context

        :param context
        :param filters

        Returns IPSec site conns
        """
        return self._get_ipsec_site2site_contexts(context, filters)

    def _get_ipsec_site2site_contexts(self, context, filters=None):
        """ Get ipsec site to site context
        :param filters
        e.g {'tenant_id': <value>,
             'vpnservice_id': <value>,
             'siteconn_id': <value>
            }
        'tenant_id' - To get s2s conns of that tenant
        'vpnservice_id' - To get s2s conns of that vpn service
        'siteconn_id' - To get a specific s2s conn

        Returns: vpnservices
            e.g { 'vpnserviceid':
                    { 'service': <VPNService>,
                  'siteconns':[ {
                                'connection': <IPSECsiteconnections>,
                                'ikepolicy': <IKEPolicy>,
                                'ipsecpolicy': <IPSECPolicy>
                                }
                              ]
                    }
                }
        """
        if not filters:
            filters = {}

        service_info = context['service_info']
        vpnservices = {}
        s_filters = {}
        if 'tenant_id' in filters:
            s_filters['tenant_id'] = [filters['tenant_id']]
        if 'vpnservice_id' in filters:
            s_filters['vpnservice_id'] = [filters['vpnservice_id']]
        if 'siteconn_id' in filters:
            s_filters['id'] = [filters['siteconn_id']]
        if 'peer_address' in filters:
            s_filters['peer_address'] = [filters['peer_address']]

        ipsec_site_conns = self.apply_filter(service_info['ipsec_site_conns'],
                                             s_filters)

        for conn in ipsec_site_conns:

            vpnservice = [vpn for vpn in service_info['vpnservices']
                          if vpn['id'] == conn['vpnservice_id']][0]

            ikepolicy = [ikepolicy for ikepolicy in service_info['ikepolicies']
                         if ikepolicy['id'] == conn['ikepolicy_id']][0]

            ipsecpolicy = [ipsecpolicy for ipsecpolicy in
                           service_info['ipsecpolicies']
                           if ipsecpolicy['id'] == conn['ipsecpolicy_id']][0]
            """
            Get the local subnet cidr
            """

            siteconn = {}
            siteconn['connection'] = conn
            siteconn['ikepolicy'] = ikepolicy
            siteconn['ipsecpolicy'] = ipsecpolicy
            vpnserviceid = vpnservice['id']

            if vpnserviceid not in vpnservices.keys():
                vpnservices[vpnserviceid] = \
                    {'service': vpnservice, 'siteconns': []}

            vpnservices[vpnserviceid]['siteconns'].append(siteconn)

        site2site_context = self._make_vpnservice_context(vpnservices)
        return site2site_context

    def _make_vpnservice_context(self, vpnservices):
        """Generate vpnservice context from the dictionary of vpnservices.
        See, if some values are not needed by agent-driver, do not pass them.
        As of now, passing everything.
        """

        return vpnservices.values()

    def _get_logical_device(self, context, filters):
        """ Get logical device from context after applying filter.
            Logical device here means pool and its related resources like vip,
            member,hm etc

        :param context
        e.g context = {'service_info': { 'pools': [pools],
                                         'members': [members],
                                         'vips': [vips],
                                         'health_monitors': [health_monitors],
                                         'ports': [ports],
                                         'subnets': [subnets],
                                       }
                      }
        :param filters e.g {'pool_id': pool_id}

        Returns: logical divice

        """
        service_info = context['service_info']
        pool_id = filters.get('pool_id')
        pool = self.get_record(service_info['pools'], 'id', pool_id)

        retval = {}
        retval['pool'] = pool  # self._make_pool_dict(pool)

        if 'vip_id' in pool and pool['vip_id'] is not None:
            vip = self.get_record(
                service_info['vips'], 'id', pool['vip_id'])
            retval['vip'] = vip  # self._make_vip_dict(vip)

        pool_members = pool['members']
        retval['members'] = []

        for pm in pool_members:
            member = self.get_record(service_info['members'], 'id', pm)
            if (member['status'] in constants.ACTIVE_PENDING_STATUSES or
                    member['status'] == constants.INACTIVE):
                retval['members'].append(member)

        pool_health_monitors = pool['health_monitors_status']
        retval['healthmonitors'] = []

        for phm in pool_health_monitors:
            if phm['status'] in constants.ACTIVE_PENDING_STATUSES:
                health_monitor = self.get_record(
                    service_info['health_monitors'],
                    'id', phm['monitor_id'])
                retval['healthmonitors'].append(health_monitor)

        retval['driver'] = pool['provider']
        return retval



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import sqlalchemy as sa

from sqlalchemy import orm
from sqlalchemy import sql

from neutron.api.v2 import attributes as attr
from neutron.common import exceptions as n_exc
from neutron.db import db_base_plugin_v2
from neutron.db import model_base
from neutron.db import models_v2


class ImplicitSubnetpool(model_base.BASEV2):
    __tablename__ = "implicit_subnetpools"
    subnetpool_id = sa.Column(sa.String(36),
                              sa.ForeignKey('subnetpools.id',
                                            ondelete="CASCADE"),
                              primary_key=True)
    is_implicit = sa.Column(sa.Boolean, nullable=False,
                            server_default=sql.false())
    subnetpool = orm.relationship(
        models_v2.SubnetPool,
        backref=orm.backref("implicit",
                            lazy="joined", cascade="delete"))


class ImplicitSubnetpoolMixin(object):
    """Mixin class for implicit subnetpool."""

    def get_implicit_subnetpool_id(self, context, tenant=None, ip_version="4"):
        pool = self.get_implicit_subnetpool(context, tenant=tenant,
                                            ip_version=ip_version)
        return pool['id'] if pool else None

    def get_implicit_subnetpool(self, context, tenant=None, ip_version="4"):
        pools = self._get_implicit_subnetpools(context, tenant=tenant,
                                               ip_version=ip_version)
        return pools[0] if pools else None

    def _get_implicit_subnetpools(self, context, tenant=None, ip_version="4"):
        admin_context = context.elevated()
        filters = {"is_implicit": [True],
                   "ip_version": [ip_version]}
        if tenant:
            filters["tenant_id"] = [tenant]
        else:
            filters["shared"] = [True]
        with context.session.begin(subtransactions=True):
            return self.get_subnetpools(admin_context, filters)

    def _get_implicit_subnetpool(self, context, subnetpool_id):
        return (context.session.query(ImplicitSubnetpool).
                filter_by(subnetpool_id=subnetpool_id)).first()

    def _subnetpool_model_hook(self, context, original_model, query):
        query = query.outerjoin(ImplicitSubnetpool,
                                (original_model.id ==
                                 ImplicitSubnetpool.subnetpool_id))
        return query

    def _subnetpool_filter_hook(self, context, original_model, conditions):
        return conditions

    def _subnetpool_result_filter_hook(self, query, filters):
        vals = filters and filters.get('is_implicit', [])
        if not vals:
            return query
        return query.filter(
            (ImplicitSubnetpool.is_implicit.in_(vals)))

    db_base_plugin_v2.NeutronDbPluginV2.register_model_query_hook(
        models_v2.SubnetPool,
        "implicit_subnetpool",
        '_subnetpool_model_hook',
        '_subnetpool_filter_hook',
        '_subnetpool_result_filter_hook')

    def _extend_subnetpool_dict_implicit(self, subnetpool_res,
                                         subnetpool_db):
        try:
            subnetpool_res["is_implicit"] = (
                subnetpool_db.implicit[0].is_implicit)
        except (IndexError, AttributeError):
            # is_implicit is not created yet when subnetpool is first added
            # to the database
            pass
        return subnetpool_res

    # Register dict extend functions for ports
    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
        attr.SUBNETPOOLS, ['_extend_subnetpool_dict_implicit'])

    def update_implicit_subnetpool(self, context, subnetpool):
        is_implicit = False
        if attr.is_attr_set(subnetpool.get('is_implicit')):
            is_implicit = subnetpool['is_implicit']
        with context.session.begin(subtransactions=True):
            if is_implicit:
                # Verify feasibility. Only one implicit SP must exist per
                # tenant (or global)
                msg = _('There can be at most one implicit '
                        'subnetpool per address family per tenant.')
                self._validate_implicit_subnetpool(
                    context, subnetpool['id'], tenant=subnetpool['tenant_id'],
                    msg=msg, ip_version=subnetpool['ip_version'])
                if subnetpool['shared']:
                    # Check globally too
                    msg = _('There can be at most one global implicit '
                            'subnetpool per address family.')
                    self._validate_implicit_subnetpool(
                        context, subnetpool['id'],
                        tenant=None,
                        msg=msg, ip_version=subnetpool['ip_version'])
            db_obj = self._get_implicit_subnetpool(
                context, subnetpool['id'])
            if db_obj:
                db_obj.is_implicit = is_implicit
            db_obj = db_obj or ImplicitSubnetpool(
                subnetpool_id=subnetpool['id'],
                is_implicit=is_implicit)
            context.session.add(db_obj)
        return is_implicit

    def _validate_implicit_subnetpool(self, context, subnetpool_id,
                                      tenant=None, msg=None, ip_version="4"):
        current_implicit_sp = self._get_implicit_subnetpools(
            context, tenant=tenant, ip_version=ip_version)
        if len(current_implicit_sp) > 1:
            raise n_exc.BadRequest(resource='subnetpools', msg=msg)
        if (len(current_implicit_sp) == 1 and
                current_implicit_sp[0]['id'] != subnetpool_id):
            raise n_exc.BadRequest(resource='subnetpools', msg=msg)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import copy

from gbpservice.contrib.nfp.config_orchestrator.common import common
from gbpservice.nfp.common import constants as const
from gbpservice.nfp.common import data_formatter as df
from gbpservice.nfp.core import context as module_context
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.lib import transport

from neutron._i18n import _LI
from neutron_lbaas.db.loadbalancer import loadbalancer_db

from oslo_log import helpers as log_helpers
import oslo_messaging as messaging

LOG = nfp_logging.getLogger(__name__)

"""
RPC handler for Loadbalancer service
"""


class LbAgent(loadbalancer_db.LoadBalancerPluginDb):
    RPC_API_VERSION = const.LOADBALANCER_RPC_API_VERSION
    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, conf, sc):
        super(LbAgent, self).__init__()
        self._conf = conf
        self._sc = sc
        self._db_inst = super(LbAgent, self)

    def _get_pools(self, **kwargs):
        context = kwargs.get('context')
        filters = {'tenant_id': [kwargs.get('tenant_id')],
                   'id': [kwargs.get('pool_id')]}
        args = {'context': context, 'filters': filters}
        pools = self._db_inst.get_pools(**args)
        for pool in pools:
            pool['description'] = kwargs.get('description')
        return pools

    def _get_vips(self, **kwargs):
        context = kwargs.get('context')
        filters = {'tenant_id': [kwargs.get('tenant_id')],
                   'pool_id': [kwargs.get('pool_id')]}
        args = {'context': context, 'filters': filters}
        vips = self._db_inst.get_vips(**args)
        for vip in vips:
            vip['description'] = kwargs.get('description')
        return vips

    def _get_members(self, **kwargs):
        context = kwargs.get('context')
        filters = {'tenant_id': [kwargs.get('tenant_id')],
                   'pool_id': [kwargs.get('pool_id')]}
        args = {'context': context, 'filters': filters}
        members = self._db_inst.get_members(**args)
        for member in members:
            member.update({'description': kwargs.get('description')})
        return members

    def _get_health_monitors(self, **kwargs):
        context = kwargs.get('context')
        filters = {'tenant_id': [kwargs.get('tenant_id')],
                   'pool_id': [kwargs.get('pool_id')]}
        args = {'context': context, 'filters': filters}
        health_monitors = self._db_inst.get_health_monitors(**args)
        for health_monitor in health_monitors:
            health_monitor.update({'description': kwargs.get('description')})
        return health_monitors

    def _get_lb_context(self, **kwargs):
        pools = self._get_pools(**kwargs)
        vips = self._get_vips(**kwargs)
        members = self._get_members(**kwargs)
        healthmonitors = self._get_health_monitors(**kwargs)
        return {'pools': pools,
                'vips': vips,
                'members': members,
                'health_monitors': healthmonitors}

    def _context(self, **kwargs):
        context = kwargs.get('context')
        if context.is_admin:
            kwargs['tenant_id'] = context.tenant_id
        lb_db = self._get_lb_context(**kwargs)
        return lb_db

    def _prepare_resource_context_dicts(self, **kwargs):
        # Prepare context_dict
        context = kwargs.get('context')
        context_resource_data = kwargs.pop('context_resource_data')
        ctx_dict = context.to_dict()
        # Collecting db entry required by configurator.
        # Addind service_info to neutron context and sending
        # dictionary format to the configurator.
        db = self._context(**kwargs)
        rsrc_ctx_dict = copy.deepcopy(ctx_dict)
        rsrc_ctx_dict.update({'service_info': db})
        rsrc_ctx_dict.update({'resource_data': context_resource_data})
        return ctx_dict, rsrc_ctx_dict

    def _get_resource_data(self, description, resource_type):
        resource_data = df.get_network_function_info(description,
                                                     resource_type)
        return resource_data

    def _update_request_data(self, body, description):
        pass

    def _data_wrapper(self, context, tenant_id, name, reason, nf, **kwargs):
        nfp_context = {}
        description = ast.literal_eval((nf['description'].split('\n'))[1])
        description.update({'tenant_id': tenant_id})
        # REVISIT(dpak): We need to avoid resource description
        # dependency in OTC and instead use neutron context description.
        context_resource_data = self._get_resource_data(description,
                                                        const.LOADBALANCER)
        if name.lower() == 'pool_health_monitor':
            pool_id = kwargs.get('pool_id')
            kwargs['health_monitor'].update({'description': str(description)})
        elif name.lower() == 'pool':
            pool_id = kwargs['pool'].get('id')
            kwargs['pool']['description'] = str(description)
        elif name.lower() == 'vip':
            pool_id = kwargs['vip'].get('pool_id')
            kwargs['vip']['description'] = str(description)
            nfp_context = {'network_function_id': nf['id'],
                           'vip_id': kwargs['vip']['id']}
        else:
            kwargs[name.lower()].update({'description': str(description)})
            pool_id = kwargs[name.lower()].get('pool_id')

        args = {'tenant_id': tenant_id,
                'pool_id': pool_id,
                'context': context,
                'description': str(description),
                'context_resource_data': context_resource_data}
        ctx_dict, rsrc_ctx_dict = self._prepare_resource_context_dicts(
            **args)

        nfp_context.update({'neutron_context': ctx_dict,
                            'requester': 'nas_service',
                            'logging_context':
                                module_context.get()['log_context']})
        resource_type = 'loadbalancer'
        resource = name
        resource_data = {'neutron_context': rsrc_ctx_dict}
        resource_data.update(**kwargs)
        body = common.prepare_request_data(nfp_context, resource,
                                           resource_type, resource_data,
                                           description['service_vendor'])
        self._update_request_data(body, description)
        return body

    def _post(self, context, tenant_id, name, nf, **kwargs):
        body = self._data_wrapper(context, tenant_id, name,
                                  'CREATE', nf, **kwargs)
        transport.send_request_to_configurator(self._conf,
                                               context, body, "CREATE")

    def _put(self, context, tenant_id, name, nf, **kwargs):
        body = self._data_wrapper(context, tenant_id, name,
                                  'UPDATE', nf, **kwargs)
        transport.send_request_to_configurator(self._conf,
                                               context, body, "UPDATE")

    def _delete(self, context, tenant_id, name, nf, **kwargs):
        body = self._data_wrapper(context, tenant_id, name,
                                  'DELETE', nf, **kwargs)
        transport.send_request_to_configurator(self._conf,
                                               context, body, "DELETE")

    def _get_pool(self, context, pool_id):
        pool = None
        try:
            pool = self._db_inst.get_pool(context, pool_id)
        except Exception as e:
            msg = ("%s" % (e))
            LOG.error(msg)
        return pool

    def _fetch_nf_from_resource_desc(self, desc):
        desc_dict = ast.literal_eval(desc)
        nf_id = desc_dict['network_function_id']
        return nf_id

    @log_helpers.log_method_call
    def create_vip(self, context, vip):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC CREATE VIP for VIP:%(vip)s"),
                 {'vip': vip})
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(vip["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._post(context, vip['tenant_id'], 'vip', nf, vip=vip)

    @log_helpers.log_method_call
    def update_vip(self, context, old_vip, vip):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC UPDATE VIP for VIP:%(vip)s "
                     "and OLD_VIP:%(old_vip)s"),
                 {'vip': vip,
                  'old_vip': old_vip})
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(vip["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._put(context, vip['tenant_id'], 'vip', nf, old_vip=old_vip,
                  vip=vip)

    @log_helpers.log_method_call
    def delete_vip(self, context, vip):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC DELETE VIP for VIP:%(vip)s"),
                 {'vip': vip})
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(vip["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._delete(context, vip['tenant_id'], 'vip', nf, vip=vip)

    @log_helpers.log_method_call
    def create_pool(self, context, pool, driver_name):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC CREATE POOL for Pool:%(pool)s"),
                 {'pool': pool})
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(pool["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._post(
            context, pool['tenant_id'],
            'pool', nf, pool=pool, driver_name=driver_name)

    @log_helpers.log_method_call
    def update_pool(self, context, old_pool, pool):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC UPDATE POOL with Pool: %(pool)s "
                     "and Old_Pool:%(old_pool)s"),
                 {'pool': pool,
                  'old_pool': old_pool})
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(pool["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._put(context, pool['tenant_id'], 'pool', nf, old_pool=old_pool,
                  pool=pool)

    @log_helpers.log_method_call
    def delete_pool(self, context, pool):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC DELETE POOL for Pool:%(pool)s"),
                 {'pool': pool})
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(pool["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._delete(context, pool['tenant_id'], 'pool', nf, pool=pool)

    @log_helpers.log_method_call
    def create_member(self, context, member):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC CREATE MEMBER for Member:%(member)s"),
                 {'member': member})
        # Fetch pool from pool_id
        pool = self._get_pool(context, member['pool_id'])
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(pool["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._post(context, member['tenant_id'], 'member', nf, member=member)

    @log_helpers.log_method_call
    def update_member(self, context, old_member, member):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC UPDATE MEMBER with Member:%(member)s "
                     "and Old_Member:%(old_member)s"),
                 {'member': member,
                  'old_member': old_member})
        # Fetch pool from pool_id
        pool = self._get_pool(context, member['pool_id'])
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(pool["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._put(context, member['tenant_id'], 'member', nf,
                  old_member=old_member, member=member)

    @log_helpers.log_method_call
    def delete_member(self, context, member):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC DELETE MEMBER for Member:%(member)s"),
                 {'member': member})
        # Fetch pool from pool_id
        pool = self._get_pool(context, member['pool_id'])
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(pool["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._delete(
            context, member['tenant_id'], 'member',
            nf, member=member)

    @log_helpers.log_method_call
    def create_pool_health_monitor(self, context, health_monitor, pool_id):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC CREATE POOL HEALTH MONITOR for HM:%(hm)s"),
                 {'hm': health_monitor})
        # Fetch pool from pool_id
        pool = self._get_pool(context, pool_id)
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(pool["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._post(context, health_monitor[
            'tenant_id'], 'pool_health_monitor',
            nf, health_monitor=health_monitor, pool_id=pool_id)

    @log_helpers.log_method_call
    def update_pool_health_monitor(self, context, old_health_monitor,
                                   health_monitor, pool_id):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC UPDATE POOL HEALTH MONITOR with "
                     "HM:%(hm)s and Old_HM:%(old_hm)s"),
                 {'hm': health_monitor,
                  'old_hm': old_health_monitor})
        # Fetch pool from pool_id
        pool = self._get_pool(context, pool_id)
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(pool["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._put(context, health_monitor[
            'tenant_id'], 'pool_health_monitor',
            nf, old_health_monitor=old_health_monitor,
            health_monitor=health_monitor, pool_id=pool_id)

    @log_helpers.log_method_call
    def delete_pool_health_monitor(self, context, health_monitor, pool_id):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC DELETE POOL HEALTH MONITOR for HM:%(hm)s"),
                 {'hm': health_monitor})
        # Fetch pool from pool_id
        pool = self._get_pool(context, pool_id)
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(pool["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._delete(
            context, health_monitor['tenant_id'], 'pool_health_monitor',
            nf, health_monitor=health_monitor, pool_id=pool_id)



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# revision identifiers, used by Alembic.
revision = '3791adbf0045'
down_revision = '2f3834ea746b'


from alembic import op
import sqlalchemy as sa


def upgrade():

    op.add_column(
        'sc_nodes',
        sa.Column('shared', sa.Boolean)
    )

    op.add_column(
        'sc_specs',
        sa.Column('shared', sa.Boolean)
    )


def downgrade():
    op.drop_column('sc_nodes', 'shared')
    op.drop_column('sc_specs', 'shared')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from neutron._i18n import _LI
from neutron.agent import securitygroups_rpc
from neutron.api import extensions
from neutron.quota import resource_registry
from neutron.scheduler import l3_agent_scheduler

from gbpservice.network.neutronv2 import local_api


# The following is to avoid excessive logging in the UTs
extensions._LW = extensions._LI
l3_agent_scheduler._LW = _LI
securitygroups_rpc._LW = securitygroups_rpc._LI
resource_registry._LW = resource_registry._LI
local_api._LW = _LI
extensions.LOG.warning = extensions.LOG.info
resource_registry.LOG.warning = resource_registry.LOG.info
l3_agent_scheduler.LOG.warning = l3_agent_scheduler.LOG.info
securitygroups_rpc.LOG.warning = securitygroups_rpc.LOG.info
local_api.LOG.warning = local_api.LOG.info



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron._i18n import _LE
from neutron._i18n import _LI
from neutron.api.v2 import attributes
from neutron.db import api as db_api
from neutron.db import db_base_plugin_v2
from neutron.db import models_v2
from neutron.db import securitygroups_db
from neutron.extensions import address_scope as as_ext
from neutron.plugins.ml2.common import exceptions as ml2_exc
from neutron.plugins.ml2 import managers as ml2_managers
from neutron.plugins.ml2 import plugin as ml2_plugin
from neutron.quota import resource_registry
from oslo_config import cfg
from oslo_log import log
from oslo_utils import excutils
from pecan import util as p_util
import six
from sqlalchemy import inspect

from gbpservice.neutron.db import implicitsubnetpool_db
from gbpservice.neutron.plugins.ml2plus import driver_context
from gbpservice.neutron.plugins.ml2plus import managers
from gbpservice.neutron.plugins.ml2plus import patch_neutron  # noqa

LOG = log.getLogger(__name__)


opts = [
    cfg.BoolOpt('refresh_network_db_obj',
                default=False,
                help=_("Refresh the network DB object to correctly "
                       "reflect the most recent state of all its "
                       "attributes. This refresh will be performed "
                       "in the _ml2_md_extend_network_dict method "
                       "inside the ml2plus plugin. The refresh option "
                       "may have a significant performace impact "
                       "and should be avoided. Hence this configuration "
                       "is set to False by default.")),
    cfg.BoolOpt('refresh_port_db_obj',
                default=False,
                help=_("Refresh the port DB object to correctly "
                       "reflect the most recent state of all its "
                       "attributes. This refresh will be performed "
                       "in the _ml2_md_extend_port_dict method "
                       "inside the ml2plus plugin. The refresh option "
                       "may have a significant performace impact "
                       "and should be avoided. Hence this configuration "
                       "is set to False by default.")),
    cfg.BoolOpt('refresh_subnet_db_obj',
                default=False,
                help=_("Refresh the subnet DB object to correctly "
                       "reflect the most recent state of all its "
                       "attributes. This refresh will be performed "
                       "in the _ml2_md_extend_subnet_dict method "
                       "inside the ml2plus plugin. The refresh option "
                       "may have a significant performace impact "
                       "and should be avoided. Hence this configuration "
                       "is set to False by default.")),
    cfg.BoolOpt('refresh_subnetpool_db_obj',
                default=False,
                help=_("Refresh the subnetpool DB object to correctly "
                       "reflect the most recent state of all its "
                       "attributes. This refresh will be performed "
                       "in the _ml2_md_extend_subnetpool_dict method "
                       "inside the ml2plus plugin. The refresh option "
                       "may have a significant performace impact "
                       "and should be avoided. Hence this configuration "
                       "is set to False by default.")),
    cfg.BoolOpt('refresh_address_scope_db_obj',
                default=False,
                help=_("Refresh the address_scope DB object to correctly "
                       "reflect the most recent state of all its "
                       "attributes. This refresh will be performed "
                       "in the _ml2_md_extend_address_scope_dict method "
                       "inside the ml2plus plugin. The refresh option "
                       "may have a significant performace impact "
                       "and should be avoided. Hence this configuration "
                       "is set to False by default.")),
]

cfg.CONF.register_opts(opts, "ml2plus")


# Copied from newton version of neutron/db/api.py.
def retry_if_session_inactive(context_var_name='context'):
    """Retries only if the session in the context is inactive.

    Calls a retry_db_errors wrapped version of the function if the context's
    session passed in is inactive, otherwise it just calls the function
    directly. This is useful to avoid retrying things inside of a transaction
    which is ineffective for DB races/errors.

    This should be used in all cases where retries are desired and the method
    accepts a context.
    """
    def decorator(f):
        try:
            # NOTE(kevinbenton): we use pecan's util function here because it
            # deals with the horrors of finding args of already decorated
            # functions
            ctx_arg_index = p_util.getargspec(f).args.index(context_var_name)
        except ValueError:
            raise RuntimeError(_LE("Could not find position of var %s")
                               % context_var_name)
        f_with_retry = db_api.retry_db_errors(f)

        @six.wraps(f)
        def wrapped(*args, **kwargs):
            # only use retry wrapper if we aren't nested in an active
            # transaction
            if context_var_name in kwargs:
                context = kwargs[context_var_name]
            else:
                context = args[ctx_arg_index]
            method = f if context.session.is_active else f_with_retry
            return method(*args, **kwargs)
        return wrapped
    return decorator


class Ml2PlusPlugin(ml2_plugin.Ml2Plugin,
                    implicitsubnetpool_db.ImplicitSubnetpoolMixin):

    """Extend the ML2 core plugin with missing functionality.

    The standard ML2 core plugin in Neutron is missing a few features
    needed for optimal APIC AIM support. This class adds those
    features, while maintaining compatibility with all standard ML2
    drivers and configuration. The only change necessary to use
    ML2Plus is to register the ml2plus entry point instead of the ml2
    entry port as Neutron's core_plugin. Drivers that need these
    features inherit from the extended MechanismDriver and
    ExtensionDriver abstract base classes.
    """

    __native_bulk_support = True
    __native_pagination_support = True
    __native_sorting_support = True
    ml2_plugin.Ml2Plugin._supported_extension_aliases += [
        "implicit-subnetpools"]

    # Override and bypass immediate base class's __init__ in order to
    # instantate extended manager class(es).
    @resource_registry.tracked_resources(
        network=models_v2.Network,
        port=models_v2.Port,
        subnet=models_v2.Subnet,
        subnetpool=models_v2.SubnetPool,
        security_group=securitygroups_db.SecurityGroup,
        security_group_rule=securitygroups_db.SecurityGroupRule)
    def __init__(self):
        LOG.info(_LI("Ml2Plus initializing"))
        # First load drivers, then initialize DB, then initialize drivers
        self.type_manager = ml2_managers.TypeManager()
        self.extension_manager = managers.ExtensionManager()
        self.mechanism_manager = managers.MechanismManager()
        super(ml2_plugin.Ml2Plugin, self).__init__()
        self.type_manager.initialize()
        self.extension_manager.initialize()
        self.mechanism_manager.initialize()
        self._setup_dhcp()
        self._start_rpc_notifiers()
        self.add_agent_status_check(self.agent_health_check)
        self._verify_service_plugins_requirements()
        self.refresh_network_db_obj = cfg.CONF.ml2plus.refresh_network_db_obj
        self.refresh_port_db_obj = cfg.CONF.ml2plus.refresh_port_db_obj
        self.refresh_subnet_db_obj = cfg.CONF.ml2plus.refresh_subnet_db_obj
        self.refresh_subnetpool_db_obj = (
            cfg.CONF.ml2plus.refresh_subnetpool_db_obj)
        self.refresh_address_scope_db_obj = (
            cfg.CONF.ml2plus.refresh_address_scope_db_obj)
        LOG.info(_LI("Modular L2 Plugin (extended) initialization complete"))

    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
               attributes.SUBNETPOOLS, ['_ml2_md_extend_subnetpool_dict'])

    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
               as_ext.ADDRESS_SCOPES, ['_ml2_md_extend_address_scope_dict'])

    def _ml2_md_extend_network_dict(self, result, netdb):
        session = inspect(netdb).session
        with session.begin(subtransactions=True):
            if self.refresh_network_db_obj:
                # In deployment it has been observed that the subnet
                # backref is sometimes stale inside the driver's
                # extend_network_dict. The call to refresh below
                # ensures the backrefs and other attributes are
                # not stale.
                session.refresh(netdb)
            self.extension_manager.extend_network_dict(session, netdb, result)

    def _ml2_md_extend_port_dict(self, result, portdb):
        session = inspect(portdb).session
        with session.begin(subtransactions=True):
            if self.refresh_port_db_obj:
                session.refresh(portdb)
            self.extension_manager.extend_port_dict(session, portdb, result)

    def _ml2_md_extend_subnet_dict(self, result, subnetdb):
        session = inspect(subnetdb).session
        with session.begin(subtransactions=True):
            if self.refresh_subnet_db_obj:
                session.refresh(subnetdb)
            self.extension_manager.extend_subnet_dict(
                session, subnetdb, result)

    def _ml2_md_extend_subnetpool_dict(self, result, subnetpooldb):
        session = inspect(subnetpooldb).session
        with session.begin(subtransactions=True):
            if self.refresh_subnetpool_db_obj:
                session.refresh(subnetpooldb)
            self.extension_manager.extend_subnetpool_dict(
                session, subnetpooldb, result)

    def _ml2_md_extend_address_scope_dict(self, result, address_scopedb):
        session = inspect(address_scopedb).session
        with session.begin(subtransactions=True):
            if self.refresh_address_scope_db_obj:
                session.refresh(address_scopedb)
            self.extension_manager.extend_address_scope_dict(
                session, address_scopedb, result)

    # Base version does not call _apply_dict_extend_functions()
    def _make_address_scope_dict(self, address_scope, fields=None):
        res = {'id': address_scope['id'],
               'name': address_scope['name'],
               'tenant_id': address_scope['tenant_id'],
               'shared': address_scope['shared'],
               'ip_version': address_scope['ip_version']}
        self._apply_dict_extend_functions(as_ext.ADDRESS_SCOPES, res,
                                          address_scope)
        return self._fields(res, fields)

    @retry_if_session_inactive()
    def create_network(self, context, network):
        self._ensure_tenant(context, network[attributes.NETWORK])
        return super(Ml2PlusPlugin, self).create_network(context, network)

    @retry_if_session_inactive()
    def update_network(self, context, id, network):
        return super(Ml2PlusPlugin, self).update_network(context, id, network)

    @retry_if_session_inactive()
    def get_network(self, context, id, fields=None):
        return super(Ml2PlusPlugin, self).get_network(context, id, fields)

    @retry_if_session_inactive()
    def get_networks(self, context, filters=None, fields=None,
                     sorts=None, limit=None, marker=None, page_reverse=False):
        return super(Ml2PlusPlugin, self).get_networks(
            context, filters, fields, sorts, limit, marker, page_reverse)

    @retry_if_session_inactive()
    def delete_network(self, context, id):
        return super(Ml2PlusPlugin, self).delete_network(context, id)

    @retry_if_session_inactive()
    def create_network_bulk(self, context, networks):
        self._ensure_tenant_bulk(context, networks[attributes.NETWORKS],
                                 attributes.NETWORK)
        return super(Ml2PlusPlugin, self).create_network_bulk(context,
                                                              networks)

    @retry_if_session_inactive()
    def create_subnet(self, context, subnet):
        self._ensure_tenant(context, subnet[attributes.SUBNET])
        return super(Ml2PlusPlugin, self).create_subnet(context, subnet)

    @retry_if_session_inactive()
    def update_subnet(self, context, id, subnet):
        return super(Ml2PlusPlugin, self).update_subnet(context, id, subnet)

    @retry_if_session_inactive()
    def get_subnet(self, context, id, fields=None):
        return super(Ml2PlusPlugin, self).get_subnet(context, id, fields)

    @retry_if_session_inactive()
    def get_subnets(self, context, filters=None, fields=None,
                    sorts=None, limit=None, marker=None, page_reverse=False):
        return super(Ml2PlusPlugin, self).get_subnets(
            context, filters, fields, sorts, limit, marker, page_reverse)

    @retry_if_session_inactive()
    def delete_subnet(self, context, id):
        return super(Ml2PlusPlugin, self).delete_subnet(context, id)

    @retry_if_session_inactive()
    def create_subnet_bulk(self, context, subnets):
        self._ensure_tenant_bulk(context, subnets[attributes.SUBNETS],
                                 attributes.SUBNET)
        return super(Ml2PlusPlugin, self).create_subnet_bulk(context,
                                                             subnets)

    @retry_if_session_inactive()
    def create_port(self, context, port):
        self._ensure_tenant(context, port[attributes.PORT])
        return super(Ml2PlusPlugin, self).create_port(context, port)

    @retry_if_session_inactive()
    def create_port_bulk(self, context, ports):
        self._ensure_tenant_bulk(context, ports[attributes.PORTS],
                                 attributes.PORT)
        return super(Ml2PlusPlugin, self).create_port_bulk(context,
                                                           ports)

    @retry_if_session_inactive()
    def update_port(self, context, id, port):
        return super(Ml2PlusPlugin, self).update_port(context, id, port)

    @retry_if_session_inactive()
    def delete_port(self, context, id, l3_port_check=True):
        return super(Ml2PlusPlugin, self).delete_port(
            context, id, l3_port_check=l3_port_check)

    @retry_if_session_inactive(context_var_name='plugin_context')
    def get_bound_port_context(self, plugin_context, port_id, host=None,
                               cached_networks=None):
        return super(Ml2PlusPlugin, self).get_bound_port_context(
            plugin_context, port_id, host, cached_networks)

    @retry_if_session_inactive()
    def update_port_status(self, context, port_id, status, host=None,
                           network=None):
        return super(Ml2PlusPlugin, self).update_port_status(
            context, port_id, status, host, network)

    @retry_if_session_inactive()
    def port_bound_to_host(self, context, port_id, host):
        return super(Ml2PlusPlugin, self).port_bound_to_host(
            context, port_id, host)

    @retry_if_session_inactive()
    def get_ports_from_devices(self, context, devices):
        return super(Ml2PlusPlugin, self).get_ports_from_devices(
            context, devices)

    @retry_if_session_inactive()
    def create_subnetpool(self, context, subnetpool):
        self._ensure_tenant(context, subnetpool[attributes.SUBNETPOOL])
        session = context.session
        with session.begin(subtransactions=True):
            result = super(Ml2PlusPlugin, self).create_subnetpool(context,
                                                                  subnetpool)
            self._update_implicit_subnetpool(context, subnetpool, result)
            self.extension_manager.process_create_subnetpool(
                context, subnetpool[attributes.SUBNETPOOL], result)
            mech_context = driver_context.SubnetPoolContext(
                self, context, result)
            self.mechanism_manager.create_subnetpool_precommit(mech_context)
        try:
            self.mechanism_manager.create_subnetpool_postcommit(mech_context)
        except ml2_exc.MechanismDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("mechanism_manager.create_subnetpool_postcommit "
                              "failed, deleting subnetpool '%s'"),
                          result['id'])
                self.delete_subnetpool(context, result['id'])
        return result

    # REVISIT(rkukura): Is create_subnetpool_bulk() needed?

    @retry_if_session_inactive()
    def update_subnetpool(self, context, id, subnetpool):
        session = context.session
        with session.begin(subtransactions=True):
            original_subnetpool = super(Ml2PlusPlugin, self).get_subnetpool(
                context, id)
            updated_subnetpool = super(Ml2PlusPlugin, self).update_subnetpool(
                context, id, subnetpool)
            self._update_implicit_subnetpool(context, subnetpool,
                                             updated_subnetpool)
            self.extension_manager.process_update_subnetpool(
                context, subnetpool[attributes.SUBNETPOOL],
                updated_subnetpool)
            mech_context = driver_context.SubnetPoolContext(
                self, context, updated_subnetpool,
                original_subnetpool=original_subnetpool)
            self.mechanism_manager.update_subnetpool_precommit(mech_context)
        self.mechanism_manager.update_subnetpool_postcommit(mech_context)
        return updated_subnetpool

    @retry_if_session_inactive()
    def get_subnetpool(self, context, id, fields=None):
        return super(Ml2PlusPlugin, self).get_subnetpool(context, id, fields)

    @retry_if_session_inactive()
    def get_subnetpools(self, context, filters=None, fields=None,
                        sorts=None, limit=None, marker=None,
                        page_reverse=False):
        return super(Ml2PlusPlugin, self).get_subnetpools(
            context, filters, fields, sorts, limit, marker, page_reverse)

    @retry_if_session_inactive()
    def delete_subnetpool(self, context, id):
        session = context.session
        with session.begin(subtransactions=True):
            subnetpool = super(Ml2PlusPlugin, self).get_subnetpool(context, id)
            mech_context = driver_context.SubnetPoolContext(
                self, context, subnetpool)
            self.mechanism_manager.delete_subnetpool_precommit(mech_context)
            super(Ml2PlusPlugin, self).delete_subnetpool(context, id)
        self.mechanism_manager.delete_subnetpool_postcommit(mech_context)

    def _update_implicit_subnetpool(self, context, request, result):
        if attributes.is_attr_set(request['subnetpool'].get('is_implicit')):
            result['is_implicit'] = request['subnetpool']['is_implicit']
            result['is_implicit'] = (
                self.update_implicit_subnetpool(context, result))

    @retry_if_session_inactive()
    def create_address_scope(self, context, address_scope):
        self._ensure_tenant(context, address_scope[as_ext.ADDRESS_SCOPE])
        session = context.session
        with session.begin(subtransactions=True):
            result = super(Ml2PlusPlugin, self).create_address_scope(
                context, address_scope)
            self.extension_manager.process_create_address_scope(
                context, address_scope[as_ext.ADDRESS_SCOPE], result)
            mech_context = driver_context.AddressScopeContext(
                self, context, result)
            self.mechanism_manager.create_address_scope_precommit(
                mech_context)
        try:
            self.mechanism_manager.create_address_scope_postcommit(
                mech_context)
        except ml2_exc.MechanismDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("mechanism_manager.create_address_scope_"
                              "postcommit failed, deleting address_scope"
                              " '%s'"),
                          result['id'])
                self.delete_address_scope(context, result['id'])
        return result

    # REVISIT(rkukura): Is create_address_scope_bulk() needed?

    @retry_if_session_inactive()
    def update_address_scope(self, context, id, address_scope):
        session = context.session
        with session.begin(subtransactions=True):
            original_address_scope = super(Ml2PlusPlugin,
                                           self).get_address_scope(context, id)
            updated_address_scope = super(Ml2PlusPlugin,
                                          self).update_address_scope(
                                              context, id, address_scope)
            self.extension_manager.process_update_address_scope(
                context, address_scope[as_ext.ADDRESS_SCOPE],
                updated_address_scope)
            mech_context = driver_context.AddressScopeContext(
                self, context, updated_address_scope,
                original_address_scope=original_address_scope)
            self.mechanism_manager.update_address_scope_precommit(mech_context)
        self.mechanism_manager.update_address_scope_postcommit(mech_context)
        return updated_address_scope

    @retry_if_session_inactive()
    def get_address_scope(self, context, id, fields=None):
        return super(Ml2PlusPlugin, self).get_address_scope(
            context, id, fields)

    @retry_if_session_inactive()
    def get_address_scopes(self, context, filters=None, fields=None,
                           sorts=None, limit=None, marker=None,
                           page_reverse=False):
        return super(Ml2PlusPlugin, self).get_address_scopes(
            context, filters, fields, sorts, limit, marker, page_reverse)

    @retry_if_session_inactive()
    def delete_address_scope(self, context, id):
        session = context.session
        with session.begin(subtransactions=True):
            address_scope = super(Ml2PlusPlugin, self).get_address_scope(
                context, id)
            mech_context = driver_context.AddressScopeContext(
                self, context, address_scope)
            self.mechanism_manager.delete_address_scope_precommit(mech_context)
            super(Ml2PlusPlugin, self).delete_address_scope(context, id)
        self.mechanism_manager.delete_address_scope_postcommit(mech_context)

    def _ensure_tenant(self, context, resource):
        tenant_id = resource['tenant_id']
        self.mechanism_manager.ensure_tenant(context, tenant_id)

    def _ensure_tenant_bulk(self, context, resources, singular):
        tenant_ids = [resource[singular]['tenant_id']
                      for resource in resources]
        for tenant_id in set(tenant_ids):
            self.mechanism_manager.ensure_tenant(context, tenant_id)

    def _get_subnetpool_id(self, context, subnet):
        # Check for regular subnetpool ID first, then Tenant's implicit,
        # then global implicit.
        ip_version = subnet['ip_version']
        return (
            super(Ml2PlusPlugin, self)._get_subnetpool_id(context, subnet) or
            self.get_implicit_subnetpool_id(context,
                                            tenant=subnet['tenant_id'],
                                            ip_version=ip_version) or
            self.get_implicit_subnetpool_id(context, tenant=None,
                                            ip_version=ip_version))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.nfp.common import constants as nfp_constants
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.core import module as nfp_api
from gbpservice.nfp.lib import transport as transport
from gbpservice.nfp.proxy_agent.lib import topics as a_topics

from neutron import context as n_context

import sys
import traceback

LOG = nfp_logging.getLogger(__name__)

ResourceMap = {
    'device_orch': a_topics.DEVICE_ORCH_TOPIC,
    'service_orch': a_topics.SERVICE_ORCH_TOPIC,
    'nas_service': a_topics.CONFIG_ORCH_TOPIC
}


"""Periodic Class to pull notification from configurator"""


class PullNotification(nfp_api.NfpEventHandler):

    def __init__(self, sc, conf):
        self._sc = sc
        self._conf = conf

    def handle_event(self, ev):
        self._sc.poll_event(ev)

    def _method_handler(self, notification):
        # Method handles notification as per resource, resource_type and method
        try:
            requester = notification['info']['context']['requester']
            if requester in ResourceMap:
                topic = ResourceMap[requester]
                context = notification['info']['context']['neutron_context']
                rpc_ctx = n_context.Context.from_dict(context)
            else:
                topic = requester.lower() + '_notifications'
                rpc_ctx = n_context.get_admin_context()
            rpcClient = transport.RPCClient(topic)
            rpcClient.cctxt.cast(rpc_ctx,
                                 'network_function_notification',
                                 notification_data=notification)
        except Exception as e:
            msg = "Exception while calling notification: %s" % (e)
            LOG.error(msg)
            raise Exception(e)

    @nfp_api.poll_event_desc(event='PULL_NOTIFICATIONS',
            spacing=nfp_constants.PULL_NOTIFICATIONS_SPACING)
    def pull_notifications(self, ev):
        """Pull and handle notification from configurator."""
        notifications = transport.get_response_from_configurator(self._conf)

        if not isinstance(notifications, list):
            message = "Notfications not list, %s" % (notifications)
            LOG.error(message)

        else:
            for notification in notifications:
                if not notification:
                    message = "Receiver Response: Empty"
                    LOG.info(message)
                    continue
                try:
                    self._method_handler(notification)
                except AttributeError:
                    exc_type, exc_value, exc_traceback = sys.exc_info()
                    message = ("AttributeError while handling"
                        "message %s : %s " % (notification,
                        traceback.format_exception(
                            exc_type, exc_value, exc_traceback)))
                    LOG.error(message)

                except Exception as e:
                    exc_type, exc_value, exc_traceback = sys.exc_info()
                    message = ("Generic exception (%s)"
                        "while handling message (%s) : %s" % (
                            e, notification, traceback.format_exception(
                                exc_type, exc_value, exc_traceback)))
                    LOG.error(message)
